# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import defaultdict
from copy import deepcopy
import os
from pathlib import Path

from colcon_core.dependency_descriptor import DependencyDescriptor


class PackageDescriptor:
    """
    A descriptor for a package.

    A packages is identified by the following triplet:
    * the 'path' which must be an existing path
    * the 'type' which must be a non-empty string
    * the 'name' which must be a non-empty string

    'dependencies' are grouped by their category as `DependencyDescriptor` or
    `str`.

    Each item in 'hooks' must be a relative path in the installation space.

    The 'metadata' dictionary can store any additional information.
    """

    __slots__ = (
        'path',
        'type',
        'name',
        'dependencies',
        'hooks',
        'metadata',
    )

    def __init__(self, path):
        """
        Descriptor for a package in a specific path.

        :param str|Path path: The location of the package
        """
        self.path = Path(str(path))
        self.type = None
        self.name = None
        self.dependencies = defaultdict(set)
        # IDEA category specific hooks
        self.hooks = []
        self.metadata = {}

    @property
    def realpath(self):
        """Resolve the realpath of the package path."""
        return os.path.realpath(str(self.path))

    def identifies_package(self):
        """
        Check if the package has a path, type and name.

        :returns: True if the descriptor has a path, type, and name
        :rtype: bool
        """
        return self.path and self.type and self.name

    def get_dependencies(self, *, categories=None):
        """
        Get the dependencies for specific categories or for all categories.

        :param Iterable[str] categories: The names of the specific categories
        :returns: The dependencies
        :rtype: set[DependencyDescriptor]
        :raises AssertionError: if the package name is listed as a dependency
        """
        dependencies = set()
        if categories is None:
            categories = self.dependencies.keys()
        for category in sorted(categories):
            dependencies |= self.dependencies[category]
        assert self.name not in dependencies, \
            "The package '{self.name}' has a dependency with the same name" \
            .format_map(locals())
        return {
            (DependencyDescriptor(d)
                if not isinstance(d, DependencyDescriptor) else d)
            for d in dependencies}

    def get_recursive_dependencies(
        self, descriptors, direct_categories=None, recursive_categories=None,
    ):
        """
        Get the recursive dependencies.

        Dependencies which are not in the set of package descriptor names are
        ignored.

        :param set descriptors: The known packages to
          consider
        :param Iterable[str] direct_categories: The names of the direct
          categories
        :param Iterable[str] recursive_categories: The names of the recursive
          categories
        :returns: The dependencies
        :rtype: set[DependencyDescriptor]
        :raises AssertionError: if a package lists itself as a dependency
        """
        # the following variable only exists for faster access within the loop
        descriptors_by_name = defaultdict(set)
        for d in descriptors:
            descriptors_by_name[d.name].add(d)
        queue = self.get_dependencies(categories=direct_categories)
        dependencies = set()
        depth = 0
        while queue:
            # ignore redundant dependencies
            level_queue = queue - dependencies
            queue.clear()
            depth += 1
            for dep in level_queue:
                # ignore circular dependencies
                if dep == self.name:
                    continue
                # ignore unknown dependencies
                # explicitly allow multiple packages with the same name
                descs = descriptors_by_name[dep]
                if not descs:
                    continue
                # recursing into the same function of the dependency descriptor
                # queue recursive dependencies
                for d in descs:
                    queue |= d.get_dependencies(
                        categories=recursive_categories)
                # duplicate the descriptor and metadata and add the depth
                dep = deepcopy(dep)
                dep.metadata['depth'] = depth
                # add dependency to result set
                dependencies.add(dep)
        return dependencies

    def __hash__(self):  # noqa: D105
        return hash((self.realpath, self.type, self.name))

    def __eq__(self, other):  # noqa: D105
        return (self.realpath, self.type, self.name) == \
            (other.realpath, other.type, other.name)

    def __str__(self):  # noqa: D105
        return '{' + ', '.join(
            ['%s: %s' % (s, getattr(self, s)) for s in self.__slots__]) + '}'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobStarted
from colcon_core.event_handler.console_start_end \
    import ConsoleStartEndEventHandler
from colcon_core.subprocess import SIGINT_RESULT
from mock import patch


def test_console_start_end():
    extension = ConsoleStartEndEventHandler()

    with patch('sys.stdout') as stdout:
        with patch('sys.stderr') as stderr:
            event = JobStarted('idA')
            extension((event, None))
            assert stdout.write.call_count == 2
            stdout.write.reset_mock()

            # success
            event = JobEnded('idA', 0)
            extension((event, None))
            assert stdout.write.call_count == 2
            stdout.write.reset_mock()

            # aborted
            event = JobEnded('idA', SIGINT_RESULT)
            extension((event, None))
            assert stdout.write.call_count == 2
            stdout.write.reset_mock()

            # failure
            event = JobEnded('idA', 1)
            extension((event, None))
            assert stderr.write.call_count == 2
            stderr.write.reset_mock()

            extension(('unknown', None))
            assert stdout.write.call_count == 0
            assert stderr.write.call_count == 0



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import copy
import traceback

from colcon_core.logging import colcon_logger
from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority

logger = colcon_logger.getChild(__name__)


class PackageAugmentationExtensionPoint:
    """
    The interface for package augmentation extensions.

    A package augmentation extension adds additional information to a package
    descriptor.

    For each instance the attribute `PACKAGE_AUGMENTATION_NAME` is being set
    to the basename of the entry point registering the extension.
    """

    """The version of the package augmentation extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of package augmentation extensions."""
    PRIORITY = 100

    def augment_packages(
        self, descs, *, additional_argument_names=None
    ):
        """
        Augment the package descriptors with additional information.

        The method is intended to be overridden in a subclass.
        The default implementation invokes `augment_package` for each package
        descriptor.

        :param descs: The package descriptors
        :param additional_argument_names: A list of additional arguments to
          consider
        """
        for desc in descs:
            self.augment_package(
                desc, additional_argument_names=additional_argument_names)

    def augment_package(
        self, desc: PackageDescriptor, *, additional_argument_names=None
    ):
        """
        Augment the package descriptor with additional information.

        The method is intended to be overridden in a subclass.
        If the `augment_packages` method is being overridden and never calls
        this method it doesn't have to be implemented.

        :param desc: The package descriptor
        :param additional_argument_names: A list of additional arguments to
          consider
        """
        raise NotImplementedError()


def get_package_augmentation_extensions():
    """
    Get the available package augmentation extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_AUGMENTATION_NAME = name
    return order_extensions_by_priority(extensions)


def augment_packages(
    descs, *, additional_argument_names=None, augmentation_extensions=None
):
    """
    Augment package descriptors with additional information.

    :param descs: the packages
    :type descs: set of
      :py:class:`colcon_core.package_descriptor.PackageDescriptor`
    """
    if augmentation_extensions is None:
        augmentation_extensions = get_package_augmentation_extensions()

    for extension in augmentation_extensions.values():
        try:
            retval = extension.augment_packages(
                descs,
                additional_argument_names=additional_argument_names)
            assert retval is None, 'augment_packages() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in augmentation extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package augmentation extension '
                "'{extension.PACKAGE_AUGMENTATION_NAME}': "
                '{e}\n{exc}'.format_map(locals()))
            # skip failing extension, continue with next one


def update_descriptor(
    desc: PackageDescriptor, data: dict, *, additional_argument_names=None
):
    """
    Update the package descriptor with additional information.

    For the keys `name` and `type` the values from `data` are set on the
    descriptor attributes with the same names.
    For the key `dependencies` the values are being added to the dependencies
    in each of the following categories: `build`, `run`, and `test`.
    For the keys `build_dependencies`, `run_dependencies`, and
    `test_dependencies` the values are being added to the dependencies in the
    category with the same name.
    For the key `hooks` the values are being added to the list of hooks.

    Any key-value pair not explicitly mentioned above is being used to update
    the metadata if the key is in the list of additional argument names.
    See :function:`update_metadata` for details how the metadata is updated.

    If the additional argument names is a list with the single value `*` all
    keys not explicitly mentioned above are being used to update the metadata.

    :param desc: the package descriptor
    :param data: The dictionary with additional information
    :param additional_argument_names: A dict of option names to destination
      names or a list of argument names
    """
    dep_types = ('build', 'run', 'test')
    # transfer generic dependencies to each specific type
    if 'dependencies' in data:
        for d in data['dependencies']:
            for dep_type in dep_types:
                desc.dependencies[dep_type].add(d)
    # transfer type specific dependencies
    for dep_type in dep_types:
        key = '{dep_type}-dependencies'.format_map(locals())
        if key in data:
            for d in data[key]:
                desc.dependencies[dep_type].add(d)

    # transfer hooks
    if 'hooks' in data:
        for d in data['hooks']:
            desc.hooks.append(d)

    # transfer any other metadata
    if additional_argument_names == ['*']:
        additional_argument_names = []
        # skip any of the already explicitly handled names
        ignored_names = ['name', 'type', 'dependencies', 'hooks']
        for dep_type in dep_types:
            ignored_names.append(
                '{dep_type}-dependencies'.format_map(locals()))
        for name in data.keys():
            if name in ignored_names:
                continue
            additional_argument_names.append(name)
    if isinstance(additional_argument_names, list):
        additional_argument_names = OrderedDict([
            (name, name) for name in additional_argument_names])
    for option, dest in (additional_argument_names or {}).items():
        if option in data:
            update_metadata(desc, dest, data[option])


def update_metadata(desc, key, value):
    """
    Update the metadata of a package descriptor.

    If the key doesn't exist in the metadata yet the key-value pair is added.

    If the key exists and the existing value as well as the passed value are
    dictionaries the existing value is updated with the passed value.
    If the key exists and the existing value as well as the passed value are
    lists the existing value is extended with the passed value.
    If the key exists and the existing value as well as the passed value are
    sets the existing value is union updated with the passed value.

    Otherwise the existing value is overwritten with the passed value.
    If the types were different a warning message is logged.

    :param desc: the package descriptor
    :param key: The key
    :param value: The value
    """
    if key not in desc.metadata:
        # add value to the metadata
        # copy value to avoid changes to either of them to affect each other
        desc.metadata[key] = copy.deepcopy(value)
        return

    old_value = desc.metadata[key]
    if isinstance(old_value, dict) and isinstance(value, dict):
        # update dictionary
        old_value.update(value)
        return

    if isinstance(old_value, list) and isinstance(value, list):
        # extend list
        old_value += value
        return

    if isinstance(old_value, set) and isinstance(value, set):
        # union update set
        old_value |= value
        return

    if type(old_value) != type(value):
        logger.warning(
            "update package '{desc.name}' metadata '{key}' from value "
            "'{old_value}' to '{value}'".format_map(locals()))

    # overwrite existing value
    # copy value to avoid changes to either of them to affect each other
    desc.metadata[key] = copy.deepcopy(value)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import sys
from tempfile import TemporaryDirectory

from colcon_core import shell
from colcon_core.plugin_system import SkipExtensionException
from colcon_core.shell.sh import ShShell
import pytest

from .run_until_complete import run_until_complete


def test_extension():
    use_all_shell_extensions = shell.use_all_shell_extensions
    shell.use_all_shell_extensions = True
    try:
        with TemporaryDirectory(prefix='test_colcon_') as prefix_path:
            _test_extension(Path(prefix_path))
    finally:
        shell.use_all_shell_extensions = use_all_shell_extensions

    if sys.platform == 'win32':
        shell.use_all_shell_extensions = False
        try:
            with pytest.raises(SkipExtensionException):
                ShShell()
        finally:
            shell.use_all_shell_extensions = use_all_shell_extensions


def _test_extension(prefix_path):
    extension = ShShell()

    # create_prefix_script
    extension.create_prefix_script(prefix_path, False)
    assert (prefix_path / 'local_setup.sh').exists()
    assert (prefix_path / '_local_setup_util_sh.py').exists()

    # create_package_script
    extension.create_package_script(
        prefix_path, 'pkg_name', [
            ('hookA.sh', '/some/path/hookA.sh'),
            ('hookB.other', '/some/path/hookB.other')])
    assert (prefix_path / 'share' / 'pkg_name' / 'package.sh').exists()
    content = (prefix_path / 'share' / 'pkg_name' / 'package.sh').read_text()
    assert 'hookA' in content
    assert 'hookB' not in content

    # create_hook_prepend_value
    hook_path = extension.create_hook_prepend_value(
        'env_hook_name', prefix_path, 'pkg_name', 'NAME', 'subdirectory')
    assert hook_path.exists()
    assert hook_path.name == 'env_hook_name.sh'
    content = hook_path.read_text()
    assert 'NAME' in content

    # generate_command_environment
    if sys.platform == 'win32':
        with pytest.raises(SkipExtensionException) as e:
            coroutine = extension.generate_command_environment(
                'task_name', prefix_path, {})
            run_until_complete(coroutine)
        assert str(e.value).endswith('Not usable on Windows systems')
    else:
        # dependency script missing
        with pytest.raises(RuntimeError) as e:
            coroutine = extension.generate_command_environment(
                'task_name', prefix_path, {'dep': str(prefix_path)})
            run_until_complete(coroutine)
        assert str(e.value) == (
            'Failed to find the following files:\n'
            '- {prefix_path}/share/dep/package.sh\n'
            'Check that the following packages have been built:\n'
            '- dep'.format_map(locals()))

        # dependency script exists
        dep_script = prefix_path / 'share' / 'dep' / 'package.sh'
        os.makedirs(str(dep_script.parent))
        dep_script.write_text('')
        coroutine = extension.generate_command_environment(
            'task_name', prefix_path, {'dep': str(prefix_path)})
        env = run_until_complete(coroutine)
        assert isinstance(env, dict)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core import location
from colcon_core.location import _create_symlink
from colcon_core.location import create_log_path
from colcon_core.location import get_config_path
from colcon_core.location import get_log_path
from colcon_core.location import set_default_config_path
from colcon_core.location import set_default_log_path
from mock import patch
import pytest

from .environment_context import EnvironmentContext


@pytest.fixture(scope='module', autouse=True)
def reset_global_variables():
    yield
    from colcon_core import location
    assert location._log_base_path is not None
    location._config_path = None
    location._config_path_env_var = None
    location._create_log_path_called = False
    location._log_base_path = None
    location._log_base_path_env_var = None
    location._log_subdirectory = None


def test_config_path():
    # use config path
    config_path = '/some/path'.replace('/', os.sep)
    with patch('colcon_core.location.logger.info') as info:
        set_default_config_path(path=config_path)
        info.assert_called_once_with(
            "Using config path '{config_path}'".format_map(locals()))

    # use config path if environment variable is not set
    config_path_env_var = 'TEST_COLCON_CONFIG_PATH'
    with patch('colcon_core.location.logger.info') as info:
        set_default_config_path(
            path=config_path, env_var=config_path_env_var)
        info.assert_called_once_with(
            "Using config path '{config_path}'".format_map(locals()))

    # use environment variable when set
    config_path = '/other/path'.replace('/', os.sep)
    with EnvironmentContext(**{config_path_env_var: config_path}):
        assert isinstance(get_config_path(), Path)
        assert str(get_config_path()) == config_path


def test_log_path():
    # use log base path
    log_base_path = '/some/path'.replace('/', os.sep)
    set_default_log_path(base_path=log_base_path)
    assert isinstance(get_log_path(), Path)
    assert str(get_log_path().parent) == log_base_path

    # use log base path if environment variable is not set
    log_base_path_env_var = 'TEST_COLCON_LOG_BASE_PATH'
    set_default_log_path(
        base_path=log_base_path, env_var=log_base_path_env_var)
    assert isinstance(get_log_path(), Path)
    assert str(get_log_path().parent) == log_base_path
    # use explicitly passed log base path even if environment variable is set
    with EnvironmentContext(**{log_base_path_env_var: '/not/used'}):
        assert isinstance(get_log_path(), Path)
        assert str(get_log_path().parent) == log_base_path

    # use environment variable when set and no base path passed
    log_base_path = '/other/path'.replace('/', os.sep)
    set_default_log_path(
        base_path=None, env_var=log_base_path_env_var)
    with EnvironmentContext(**{log_base_path_env_var: log_base_path}):
        assert isinstance(get_log_path(), Path)
        assert str(get_log_path().parent) == log_base_path

    # use specific subdirectory
    subdirectory = 'sub'
    set_default_log_path(
        base_path=log_base_path, env_var=log_base_path_env_var,
        subdirectory=subdirectory)
    assert isinstance(get_log_path(), Path)
    assert get_log_path() == Path(log_base_path) / subdirectory


def test_create_log_path():
    # no need to create a directory if the path already exists
    log_path = Path(os.getcwd())
    set_default_log_path(
        base_path=log_path.parent, subdirectory=log_path.name)
    with patch('os.makedirs') as makedirs:
        makedirs.side_effect = AssertionError('should not be called')
        create_log_path('verb')
    # no latest symlink is being created either
    assert not (log_path.parent / 'latest').exists()

    subdirectory = 'sub'
    with TemporaryDirectory(prefix='test_colcon_') as log_path:
        log_path = Path(log_path)
        set_default_log_path(base_path=log_path, subdirectory=subdirectory)

        # repeated call is a noop
        with patch('os.makedirs') as makedirs:
            makedirs.side_effect = AssertionError('should not be called')
            create_log_path('verb2')
        assert not (log_path.parent / 'latest').exists()

        # create a directory and symlink when the path doesn't exist
        location._create_log_path_called = False
        with patch('os.makedirs', wraps=os.makedirs) as makedirs:
            create_log_path('verb')
            makedirs.assert_called_once_with(str(log_path / subdirectory))
        assert (log_path / subdirectory).exists()

        # since the directory already exists create one with a suffix
        location._create_log_path_called = False
        with patch('os.makedirs', wraps=os.makedirs) as makedirs:
            create_log_path('verb')
            assert makedirs.call_count == 2
            assert len(makedirs.call_args_list[0][0]) == 1
            assert makedirs.call_args_list[0][0][0] == str(
                log_path / subdirectory)
            assert len(makedirs.call_args_list[1][0]) == 1
            assert makedirs.call_args_list[1][0][0] == str(
                log_path / subdirectory) + '_2'
        assert (log_path / (str(subdirectory) + '_2')).exists()

        # and another increment of the suffix
        location._create_log_path_called = False
        location._log_subdirectory = subdirectory
        with patch('os.makedirs', wraps=os.makedirs) as makedirs:
            create_log_path('verb')
            assert makedirs.call_count == 3
            assert len(makedirs.call_args_list[0][0]) == 1
            assert makedirs.call_args_list[0][0][0] == str(
                log_path / subdirectory)
            assert len(makedirs.call_args_list[1][0]) == 1
            assert makedirs.call_args_list[1][0][0] == str(
                log_path / subdirectory) + '_2'
            assert len(makedirs.call_args_list[2][0]) == 1
            assert makedirs.call_args_list[2][0][0] == str(
                log_path / subdirectory) + '_3'
        assert (log_path / (str(subdirectory) + '_3')).exists()
        subdirectory += '_3'

        # check that `latest_verb` was created and points to the subdirectory
        assert (log_path / 'latest_verb').is_symlink()
        assert (log_path / 'latest_verb').resolve() == \
            (log_path / subdirectory).resolve()

        # check that `latest` was created and points to the subdirectory
        assert (log_path / 'latest').is_symlink()
        assert (log_path / 'latest').resolve() == \
            (log_path / subdirectory).resolve()

        # create directory but correct latest symlink already exists
        (log_path / subdirectory).rmdir()
        location._create_log_path_called = False
        create_log_path('verb')
        assert (log_path / subdirectory).exists()
        assert (log_path / 'latest').is_symlink()
        assert (log_path / 'latest').resolve() == \
            (log_path / subdirectory).resolve()

        # create directory and update latest symlink
        subdirectory = 'other_sub'
        set_default_log_path(base_path=log_path, subdirectory=subdirectory)
        location._create_log_path_called = False
        create_log_path('verb')
        assert (log_path / subdirectory).exists()
        assert (log_path / 'latest').is_symlink()
        assert (log_path / 'latest').resolve() == \
            (log_path / subdirectory).resolve()

        # create directory but latest is not a symlink
        (log_path / subdirectory).rmdir()
        (log_path / 'latest').unlink()
        (log_path / 'latest').mkdir()
        location._create_log_path_called = False
        create_log_path('verb')
        assert (log_path / subdirectory).exists()
        assert not (log_path / 'latest').is_symlink()


def test__create_symlink():
    # check cases where functions raise exceptions and ensure it is being
    # handled gracefully
    with TemporaryDirectory(prefix='test_colcon_') as path:
        path = Path(path)

        # relative path couldn't be computed, symlink couldn't be created
        _create_symlink(path / 'nowhere', Path('/foo/bar'))

        # unlinking symlink failed
        class DummyPath:

            def __init__(self):
                self.parent = 'parent'

            def exists(self):
                return False

            def is_symlink(self):
                return True

            def unlink(self):
                raise FileNotFoundError()

        with patch('os.symlink') as symlink:
            _create_symlink(path / 'src', DummyPath())
            assert symlink.call_count == 1

        # (Windows) OSError: symbolic link privilege not held
        class ValidPath(DummyPath):

            def is_symlink(self):
                return False

        with patch('os.symlink') as symlink:
            symlink.side_effect = OSError()
            _create_symlink(path / 'src', ValidPath())
            assert symlink.call_count == 1



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import defaultdict
import os
import traceback

from colcon_core.environment_variable import EnvironmentVariable
from colcon_core.logging import colcon_logger
from pkg_resources import iter_entry_points
from pkg_resources import WorkingSet

"""Environment variable to blacklist extensions"""
EXTENSION_BLACKLIST_ENVIRONMENT_VARIABLE = EnvironmentVariable(
    'COLCON_EXTENSION_BLACKLIST',
    'Blacklist extensions which should not be used')

logger = colcon_logger.getChild(__name__)


"""
The group name for entry points identifying colcon extension points.

While all entry points in this package start with `colcon_core.` other
distributions might define entry points with a different prefix.
Those need to be declared using this group name.
"""
EXTENSION_POINT_GROUP_NAME = 'colcon_core.extension_point'


def get_all_entry_points():
    """
    Get all entry points related to `colcon` and any of its extensions.

    :returns: mapping of entry point names to :class:`pkg_resources.EntryPoint`
      instances
    :rtype: dict
    """
    global EXTENSION_POINT_GROUP_NAME
    colcon_extension_points = get_entry_points(EXTENSION_POINT_GROUP_NAME)

    entry_points = defaultdict(dict)
    working_set = WorkingSet()
    for dist in sorted(working_set):
        entry_map = dist.get_entry_map()
        for group_name in entry_map.keys():
            # skip groups which are not registered as extension points
            if group_name not in colcon_extension_points:
                continue

            group = entry_map[group_name]
            for entry_point_name, entry_point in group.items():
                entry_point.group_name = group_name
                if entry_point_name in entry_points[group_name]:
                    previous = entry_points[group_name][entry_point_name]
                    logger.error(
                        "Entry point '{group_name}.{entry_point_name}' is "
                        "declared multiple times, '{entry_point}' overwriting "
                        "'{previous}'".format_map(locals()))
                entry_points[group_name][entry_point_name] = \
                    (dist, entry_point)
    return entry_points


def get_entry_points(group_name):
    """
    Get the entry points for a specific group.

    :param str group_name: the name of the `entry_point` group
    :returns: mapping of group names to dictionaries which map entry point
      names to :class:`pkg_resources.EntryPoint` instances
    :rtype: dict
    """
    entry_points = {}
    for entry_point in iter_entry_points(group=group_name):
        entry_point.group_name = group_name
        if entry_point.name in entry_points:
            previous_entry_point = entry_points[entry_point.name]
            logger.error(
                "Entry point '{group_name}.{entry_point.name}' is declared "
                "multiple times, '{entry_point}' overwriting "
                "'{previous_entry_point}'".format_map(locals()))
        entry_points[entry_point.name] = entry_point
    return entry_points


def load_entry_points(group_name):
    """
    Load the entry points for a specific group.

    :param str group_name: the name of the `entry_point` group
    :returns: mapping of entry point names to loaded entry points
    :rtype: dict
    """
    extension_types = {}
    for entry_point in get_entry_points(group_name).values():
        try:
            extension_type = load_entry_point(entry_point)
        except RuntimeError:
            continue
        except Exception as e:  # noqa: F841
            # catch exceptions raised when loading entry point
            exc = traceback.format_exc()
            logger.error(
                'Exception loading extension '
                "'{group_name}.{entry_point.name}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing entry point, continue with next one
            continue
        extension_types[entry_point.name] = extension_type
    return extension_types


def load_entry_point(entry_point):
    """
    Load the entry point.

    :param entry_point: the :class:`pkg_resources.EntryPoint` instance
    :returns: the loaded entry point
    :raises RuntimeError: if either the group name or the entry point name is
      listed in the environment variable
      :const:`EXTENSION_BLACKLIST_ENVIRONMENT_VARIABLE`
    """
    global EXTENSION_BLACKLIST_ENVIRONMENT_VARIABLE
    blacklist = os.environ.get(
        EXTENSION_BLACKLIST_ENVIRONMENT_VARIABLE.name, None)
    if blacklist:
        blacklist = blacklist.split(os.pathsep)
        env_var_name = EXTENSION_BLACKLIST_ENVIRONMENT_VARIABLE.name
        if entry_point.group_name in blacklist:
            raise RuntimeError(
                'The entry point group name is listed in the environment '
                "variable '{env_var_name}'".format_map(locals()))
        full_name = '{entry_point.group_name}.{entry_point.name}' \
            .format_map(locals())
        if full_name in blacklist:
            raise RuntimeError(
                'The entry point name is listed in the environment variable '
                "'{env_var_name}'".format_map(locals()))
    return entry_point.load()



# Copyright 2020 Dirk Thomas
# Licensed under the Apache License, Version 2.0


class BooleanDefaultValue:
    """Boolean value distinguishable from explicitly passed arguments."""

    def __init__(self, value):  # noqa: D107
        assert isinstance(value, bool)
        self._value = value

    def __bool__(self):  # noqa: D105
        return self._value


class ListDefaultValue(list):
    """List value distinguishable from explicitly passed arguments."""

    pass


class StringDefaultValue(str):
    """String value distinguishable from explicitly passed arguments."""

    pass


_types = {
    bool: BooleanDefaultValue,
    list: ListDefaultValue,
    str: StringDefaultValue,
}


def wrap_default_value(value):
    """
    Wrap a default value in a custom type.

    The custom type makes the default value distinguishable from explicitly
    passed arguments independent of the value.
    The custom types try to match the behavior of the original type as much as
    possible.

    :param value: The default value to be wrapped
    :returns: The wrapped value if the value type is supported, otherwise the
      passed value
    """
    global _types
    if is_default_value(value):
        raise ValueError(
            'the passed value is already wrapped: ' + str(type(value)))
    if type(value) in _types:
        return _types[type(value)](value)
    return value


def is_default_value(value):
    """Check if a value is a default value."""
    global _types
    return isinstance(value, tuple(_types.values()))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_identification import IgnoreLocationException
from colcon_core.package_identification \
    import PackageIdentificationExtensionPoint
from colcon_core.plugin_system import satisfies_version

IGNORE_MARKER = 'COLCON_IGNORE'


class IgnorePackageIdentification(PackageIdentificationExtensionPoint):
    """Ignore paths containing a `COLCON_IGNORE` file."""

    # the priority needs to be higher than all other extensions
    PRIORITY = 1000

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PackageIdentificationExtensionPoint.EXTENSION_POINT_VERSION,
            '^1.0')

    def identify(self, desc):  # noqa: D102
        colcon_ignore = desc.path / IGNORE_MARKER
        if colcon_ignore.exists():
            raise IgnoreLocationException()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import sys

from colcon_core import shell
from colcon_core.environment import EnvironmentExtensionPoint
from colcon_core.environment import logger
from colcon_core.plugin_system import satisfies_version


class PathEnvironment(EnvironmentExtensionPoint):
    """Extend the `PATH` variable to find executables."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EnvironmentExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def create_environment_hooks(self, prefix_path, pkg_name):  # noqa: D102
        hooks = self._create_environment_hooks(prefix_path, pkg_name, 'bin')
        if sys.platform == 'win32':
            hooks += self._create_environment_hooks(
                prefix_path, pkg_name, 'Scripts', '-scripts')
        return hooks

    def _create_environment_hooks(
        self, prefix_path, pkg_name, subdirectory, suffix=''
    ):
        hooks = []
        bin_path = prefix_path / subdirectory
        logger.log(1, "checking '%s'" % bin_path)
        try:
            names = os.listdir(str(bin_path))
        except FileNotFoundError:
            pass
        else:
            for name in names:
                if not (bin_path / name).is_file():
                    continue
                hooks += shell.create_environment_hook(
                    'path' + suffix, prefix_path, pkg_name, 'PATH',
                    subdirectory, mode='prepend')
                break

        return hooks



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict

from colcon_core.package_decorator import PackageDecorator


def topological_order_packages(
    descriptors, direct_categories=None, recursive_categories=None,
):
    """
    Order packages topologically.

    :param descriptors: the package descriptors
    :type descriptors: set of
      :py:class:`colcon_core.package_descriptor.PackageDescriptor`

    :returns: list of package decorators
    :rtype: list of :py:class:`colcon_core.package_decorator.PackageDecorator`
    """
    # get recursive dependencies for all packages
    queued = set()
    for descriptor in descriptors:
        rec_deps = descriptor.get_recursive_dependencies(
            descriptors,
            direct_categories=direct_categories,
            recursive_categories=recursive_categories)
        d = _PackageDependencies(
            descriptor=descriptor,
            recursive_dependencies=rec_deps,
            remaining_dependencies={d.name for d in rec_deps},
        )
        queued.add(d)

    ordered = OrderedDict()
    while len(ordered) < len(descriptors):
        # remove dependencies on already ordered packages
        ordered_names = {descriptor.name for descriptor in ordered.keys()}
        for q in queued:
            q.remaining_dependencies -= ordered_names

        # find all queued packages without remaining dependencies
        ready = list(filter(lambda q: not q.remaining_dependencies, queued))
        if not ready:
            lines = [
                '%s: %s' % (
                    q.descriptor.name, sorted(q.remaining_dependencies))
                for q in queued]
            lines.sort()
            raise RuntimeError(
                'Unable to order packages topologically:\n' + '\n'.join(lines))

        # order ready jobs alphabetically for a deterministic order
        ready.sort(key=lambda d: d.descriptor.name)

        # add all ready jobs to ordered dictionary
        for r in ready:
            ordered[r.descriptor] = r.recursive_dependencies
            queued.remove(r)

    # create ordered list of package decorators
    decorators = []
    ordered_keys = [descriptor.name for descriptor in ordered.keys()]
    for descriptor, recursive_dependencies in ordered.items():
        decorator = PackageDecorator(descriptor)
        # reorder recursive dependencies according to the topological ordering
        decorator.recursive_dependencies = sorted(
            (d for d in recursive_dependencies if d in ordered_keys),
            key=ordered_keys.index)
        decorators.append(decorator)

    return decorators


class _PackageDependencies:

    __slots__ = (
        'descriptor', 'recursive_dependencies', 'remaining_dependencies')

    def __init__(
        self, descriptor, recursive_dependencies, remaining_dependencies,
    ):
        self.descriptor = descriptor
        self.recursive_dependencies = recursive_dependencies
        self.remaining_dependencies = remaining_dependencies



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from distutils.sysconfig import get_python_lib
import os
from pathlib import Path
import shutil
import sys
from sys import executable

from colcon_core.environment import create_environment_hooks
from colcon_core.environment import create_environment_scripts
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import create_environment_hook
from colcon_core.shell import get_command_environment
from colcon_core.task import check_call
from colcon_core.task import TaskExtensionPoint
from colcon_core.task.python import get_data_files_mapping
from colcon_core.task.python import get_setup_data

logger = colcon_logger.getChild(__name__)


class PythonBuildTask(TaskExtensionPoint):
    """Build Python packages."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    async def build(self, *, additional_hooks=None):  # noqa: D102
        pkg = self.context.pkg
        args = self.context.args

        logger.info(
            "Building Python package in '{args.path}'".format_map(locals()))

        try:
            env = await get_command_environment(
                'setup_py', args.build_base, self.context.dependencies)
        except RuntimeError as e:
            logger.error(str(e))
            return 1
        setup_py_data = get_setup_data(self.context.pkg, env)

        # `setup.py egg_info` requires the --egg-base to exist
        os.makedirs(args.build_base, exist_ok=True)
        # `setup.py develop|install` requires the python lib path to exist
        python_lib = os.path.join(
            args.install_base, self._get_python_lib(args))
        os.makedirs(python_lib, exist_ok=True)
        # and being in the PYTHONPATH
        env = dict(env)
        env['PYTHONPATH'] = python_lib + os.pathsep + \
            env.get('PYTHONPATH', '')

        if not args.symlink_install:
            rc = await self._undo_develop(pkg, args, env)
            if rc and rc.returncode:
                return rc.returncode

            # invoke `setup.py install` step with lots of arguments
            # to avoid placing any files in the source space
            cmd = [
                executable, 'setup.py',
                'egg_info', '--egg-base', args.build_base,
                'build', '--build-base', os.path.join(
                    args.build_base, 'build'),
                'install', '--prefix', args.install_base,
                '--record', os.path.join(args.build_base, 'install.log'),
                # prevent installation of dependencies specified in setup.py
                '--single-version-externally-managed',
            ]
            self._append_install_layout(args, cmd)
            rc = await check_call(self.context, cmd, cwd=args.path, env=env)
            if rc and rc.returncode:
                return rc.returncode

        else:
            self._undo_install(pkg, args, setup_py_data, python_lib)
            self._symlinks_in_build(args, setup_py_data)

            # invoke `setup.py develop` step in build space
            # to avoid placing any files in the source space

            # --editable causes this to skip creating/editing the
            # easy-install.pth file
            cmd = [
                executable, 'setup.py',
                'develop', '--prefix', args.install_base,
                '--editable',
                '--build-directory', os.path.join(args.build_base, 'build'),
                '--no-deps',
            ]
            if setup_py_data.get('data_files'):
                cmd += ['install_data', '--install-dir', args.install_base]
            self._append_install_layout(args, cmd)
            rc = await check_call(
                self.context, cmd, cwd=args.build_base, env=env)
            if rc and rc.returncode:
                return rc.returncode

            # explicitly add the build directory to the PYTHONPATH
            # to maintain the desired order
            if additional_hooks is None:
                additional_hooks = []
            base_path = args.build_base
            # if the Python packages are in a subdirectory
            # that needs to be appended to the build directory
            package_dir = setup_py_data.get('package_dir') or {}
            if '' in package_dir:
                base_path = os.path.join(base_path, package_dir[''])
            additional_hooks += create_environment_hook(
                'pythonpath_develop', Path(base_path), pkg.name, 'PYTHONPATH',
                base_path, mode='prepend')

        hooks = create_environment_hooks(args.install_base, pkg.name)
        create_environment_scripts(
            pkg, args, default_hooks=hooks, additional_hooks=additional_hooks)

    async def _undo_develop(self, pkg, args, env):
        # undo previous develop if .egg-info is found and develop symlinks
        egg_info = os.path.join(
            args.build_base, '%s.egg-info' % pkg.name.replace('-', '_'))
        setup_py_build_space = os.path.join(args.build_base, 'setup.py')
        if os.path.exists(egg_info) and os.path.islink(setup_py_build_space):
            cmd = [
                executable, 'setup.py',
                'develop', '--prefix', args.install_base,
                '--uninstall', '--editable',
                '--build-directory', os.path.join(args.build_base, 'build')
            ]
            rc = await check_call(
                self.context, cmd, cwd=args.build_base, env=env)
            if rc:
                return rc

    def _undo_install(self, pkg, args, setup_py_data, python_lib):
        # undo previous install if install.log is found
        install_log = os.path.join(args.build_base, 'install.log')
        if not os.path.exists(install_log):
            return
        with open(install_log, 'r') as h:
            lines = [l.rstrip() for l in h.readlines()]

        packages = setup_py_data.get('packages') or []
        for module_name in packages:
            if module_name in sys.modules:
                logger.warning(
                    "Switching to 'develop' for package '{pkg.name}' while it "
                    'is being used might result in import errors later'
                    .format_map(locals()))
                break

        # remove previously installed files
        directories = set()
        python_lib = python_lib + os.sep
        for line in lines:
            if not os.path.exists(line):
                continue
            if not line.startswith(python_lib):
                logger.debug(
                    'While undoing a previous installation files outside the '
                    'Python library path are being ignored: {line}'
                    .format_map(locals()))
                continue
            if not os.path.isdir(line):
                os.remove(line)
                # collect all parent directories until install base
                while True:
                    line = os.path.dirname(line)
                    if not line.startswith(python_lib):
                        break
                    directories.add(line)
        # remove empty directories
        for d in sorted(directories, reverse=True):
            try:
                os.rmdir(d)
            except OSError:
                pass
        os.remove(install_log)

    def _symlinks_in_build(self, args, setup_py_data):
        items = ['setup.py']
        # add setup.cfg if available
        if os.path.exists(os.path.join(args.path, 'setup.cfg')):
            items.append('setup.cfg')
        # add all first level packages
        package_dir = setup_py_data.get('package_dir') or {}
        for package in setup_py_data.get('packages') or []:
            if '.' in package:
                continue
            if package in package_dir:
                items.append(package_dir[package])
            elif '' in package_dir:
                items.append(os.path.join(package_dir[''], package))
            else:
                items.append(package)
        # relative python-ish paths are allowed as entries in py_modules, see:
        # https://docs.python.org/3.5/distutils/setupscript.html#listing-individual-modules
        py_modules = setup_py_data.get('py_modules')
        if py_modules:
            py_modules_list = [
                p.replace('.', os.path.sep) + '.py' for p in py_modules]
            for py_module in py_modules_list:
                if not os.path.exists(os.path.join(args.path, py_module)):
                    raise RuntimeError(
                        "Provided py_modules '{py_module}' does not exist"
                        .format_map(locals()))
            items += py_modules_list
        data_files = get_data_files_mapping(
            setup_py_data.get('data_files') or [])
        for source in data_files.keys():
            # work around data files incorrectly defined as not relative
            if os.path.isabs(source):
                source = os.path.relpath(source, args.path)
            items.append(source)

        # symlink files / directories from source space into build space
        for item in items:
            src = os.path.join(args.path, item)
            dst = os.path.join(args.build_base, item)
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            if os.path.islink(dst):
                if not os.path.exists(dst) or not os.path.samefile(src, dst):
                    os.unlink(dst)
            elif os.path.isfile(dst):
                os.remove(dst)
            elif os.path.isdir(dst):
                shutil.rmtree(dst)
            if not os.path.exists(dst):
                os.symlink(src, dst)

    def _get_python_lib(self, args):
        path = get_python_lib(prefix=args.install_base)
        return os.path.relpath(path, start=args.install_base)

    def _append_install_layout(self, args, cmd):
        if 'dist-packages' in self._get_python_lib(args):
            cmd += ['--install-layout', 'deb']



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.plugin_system import get_first_line_doc
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority


class EventHandlerExtensionPoint:
    """
    The interface for event handler extensions.

    An event handler extension processes events.

    For each instance the attribute `EVENT_HANDLER_NAME` is being set to the
    basename of the entry point registering the extension.

    Custom event handlers don't need to be subclasses but only be callables
    accepting a single `event` argument.
    They are being registered as observers at an
    :py:class:`colcon_core.EventReactor` instance.
    The handler should check the type of the event and only act on known types.
    """

    """The version of the event handler extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of event handler extensions."""
    PRIORITY = 100

    def __init__(self):  # noqa: D107
        super().__init__()
        self.context = None
        self.enabled = True

    def __call__(self, event):
        """
        Process an event if the event type is known.

        This method must be overridden in a subclass.

        :param event: The event
        """
        raise NotImplementedError()


def get_event_handler_extensions(*, context):
    """
    Get the available event handler extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.EVENT_HANDLER_NAME = name
        extension.context = context
    return order_extensions_by_priority(extensions)


def add_event_handler_arguments(parser):
    """
    Add the command line arguments for the event handler extensions.

    :param parser: The argument parser
    """
    group = parser.add_argument_group(title='Event handler arguments')
    extensions = get_event_handler_extensions(context=None)

    completions = []
    defaults = []
    descriptions = ''
    for key in sorted(extensions.keys()):
        # only offer completion for non-default choices
        completions.append(key + ('-' if extensions[key].enabled else '+'))
        defaults.append(key + ('+' if extensions[key].enabled else '-'))
        extension = extensions[key]
        desc = get_first_line_doc(extension)
        # ignore extensions without a description
        # since they are already listed in the defaults
        if desc:
            # it requires a custom formatter to maintain the newline
            descriptions += '\n* {key}: {desc}'.format_map(locals())

    argument = group.add_argument(
        '--event-handlers',
        nargs='*', choices=completions + defaults,
        metavar=('name1+', 'name2-'),
        help='Enable (+) or disable (-) event handlers (default: %s)%s' %
             (' '.join(defaults), descriptions))

    def choices_completer(prefix, **kwargs):
        nonlocal completions
        return (c for c in completions if c.startswith(prefix))

    argument.completer = choices_completer


def apply_event_handler_arguments(extensions, args):
    """
    Enable/disable the event handler extensions based on the passed arguments.

    :param extensions: The event handler extensions
    :param args: The parsed command line arguments
    """
    for arg in (args.event_handlers or []):
        suffix = arg[-1]
        assert suffix in ('+', '-')
        key = arg[:-1]
        extension = extensions[key]
        extension.enabled = suffix == '+'


def format_duration(seconds, *, fixed_decimal_points=None):
    """
    Stringify a duration as hours, minutes, seconds and decimal points.

    :param float seconds: The duration in seconds
    :param int fixed_decimal_points: A fixed number of decimal points for the
      seconds, if None two, one or none will be used depending on the duration
      size
    :returns: The string representation of the duration
    :rtype: str
    """
    if seconds < 0.0:
        raise ValueError(
            "The duration '{seconds}' must be a non-negative number"
            .format_map(locals()))
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)

    # determine number of decimal points for seconds
    if fixed_decimal_points is not None:
        decimal_points = fixed_decimal_points
    elif hours or minutes:
        decimal_points = 0
    # compare rounded number to account for floating point imprecision
    elif round(seconds, 2) < 9.995:
        decimal_points = 2
    else:
        decimal_points = 1

    # check if rounding of seconds pushes it to a full minute
    if round(seconds, decimal_points) >= 60.0:
        # check if the desired decimal points change due to the wrapping
        if fixed_decimal_points is None and not minutes and not hours:
            decimal_points = 0
        seconds = 0.0
        minutes += 1

    # check if rounding of minutes pushes it to a full hour
    if round(minutes, 0) >= 60.0:
        minutes = 0.0
        hours += 1

    format_string = '{seconds:.%sf}s' % decimal_points
    if hours or minutes:
        format_string = '{minutes:.0f}min ' + format_string
        if hours:
            format_string = '{hours:.0f}h ' + format_string
    return format_string.format_map(locals())



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import re
import traceback

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import get_first_line_doc
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority
from colcon_core.plugin_system import satisfies_version
from colcon_core.shell import get_command_environment
from colcon_core.task import TaskExtensionPoint
from colcon_core.task.python import get_setup_data

logger = colcon_logger.getChild(__name__)


class PythonTestTask(TaskExtensionPoint):
    """Test Python packages."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(TaskExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        add_python_testing_step_arguments(parser)

    async def test(self, *, additional_hooks=None):  # noqa: D102
        pkg = self.context.pkg
        args = self.context.args

        logger.info(
            "Testing Python package in '{args.path}'".format_map(locals()))

        try:
            env = await get_command_environment(
                'setup_py', args.build_base, self.context.dependencies)
        except RuntimeError as e:
            logger.error(str(e))
            return 1
        setup_py_data = get_setup_data(self.context.pkg, env)

        # select the step extension which should perform the python testing
        if args.python_testing:
            key = args.python_testing
            extension = get_python_testing_step_extension(key)
        else:
            extensions = get_python_testing_step_extensions()
            for key, extension in extensions.items():
                logger.log(
                    1, "test() by extension '{key}'".format_map(locals()))
                try:
                    matched = extension.match(self.context, env, setup_py_data)
                except Exception as e:  # noqa: F841
                    # catch exceptions raised in python testing step extension
                    exc = traceback.format_exc()
                    logger.error(
                        'Exception in Python testing step extension '
                        "'{extension.STEP_TYPE}': {e}\n{exc}"
                        .format_map(locals()))
                    # skip failing extension, continue with next one
                    continue
                if matched:
                    break
            else:
                logger.warning(
                    "No Python Testing Step extension matched in '{args.path}'"
                    .format_map(locals()))
                return

        logger.log(
            1, "test.step() by extension '{key}'".format_map(locals()))
        try:
            return await extension.step(self.context, env, setup_py_data)
        except Exception as e:  # noqa: F841
            # catch exceptions raised in python testing step extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in Python testing step extension '
                "'{extension.STEP_TYPE}': {e}\n{exc}".format_map(locals()))
            return 1


class PythonTestingStepExtensionPoint:
    """
    The interface for Python testing step extensions.

    A Python testing step extension performs testing of a Python package.

    For each instance the attribute `STEP_TYPE` is being set to the basename of
    the entry point registering the extension.
    """

    """The version of the Python testing step extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of Python testing step extensions."""
    PRIORITY = 100

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the Python testing step.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def match(self, context, env, setup_py_data):
        """
        Determine if this instance claims to process the specific package.

        This method must be overridden in a subclass.

        :param context: The task context describing the package
        :param env: The environment dictionary
        :param setup_py_data: The data extracted from the setup.py file
        :returns: True if it claims to process the package, False otherwise
        :rtype: bool
        """
        raise NotImplementedError()

    async def step(self):
        """
        Execute the Python testing step logic.

        This method must be overridden in a subclass.

        :returns: The return code
        """
        raise NotImplementedError()


def get_python_testing_step_extensions():
    """
    Get the available Python testing step extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(
        'colcon_core.python_testing', unique_instance=False)
    for name in list(extensions.keys()):
        extension = extensions[name]
        extension.STEP_TYPE = name
    return order_extensions_by_priority(extensions)


def add_python_testing_step_arguments(parser):
    """
    Add the command line arguments for the Python testing step extensions.

    :param parser: The argument parser
    """
    extensions = get_python_testing_step_extensions()
    descriptions = ''
    for key, extension in extensions.items():
        desc = get_first_line_doc(extension)
        if not desc:
            # show extensions without a description
            # to mention the available options
            desc = '<no description>'
        # it requires a custom formatter to maintain the newline
        descriptions += '\n* {key}: {desc}'.format_map(locals())

    parser.add_argument(
        '--python-testing', type=str, choices=sorted(extensions.keys()),
        help='The Python testing framework to use (default: determined '
             'based on the packages `tests_require`)'
             '{descriptions}'.format_map(locals()))

    for extension in extensions.values():
        try:
            retval = extension.add_arguments(parser=parser)
            assert retval is None, 'add_arguments() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in package selection extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in Python testing step extension '
                "'{extension.STEP_TYPE}': {e}\n{exc}".format_map(locals()))
            # skip failing extension, continue with next one


def get_python_testing_step_extension(step_name):
    """
    Get a specific Python testing step extension.

    :param str step_name: The entry point name of the extension
    :returns: A unique instance of the extension, otherwise None
    """
    extensions = instantiate_extensions(
        'colcon_core.python_testing.' + step_name, unique_instance=True)
    if step_name not in extensions:
        return None
    extension = extensions[step_name]
    extension.STEP_NAME = step_name
    return extension


def has_test_dependency(setup_py_data, name):
    """
    Check if the package has a specific test dependency.

    :param dict setup_py_data: The meta information of the package
    :returns: True if the package has a test dependency on the given name,
      False otherwise
    :rtype: bool
    """
    tests_require = setup_py_data.get('tests_require', [])
    for d in tests_require:
        # the name might be followed by a version
        # separated by any of the following: ' ', <, >, <=, >=, ==, !=
        parts = re.split(r' |<|=|>|!', d)
        if parts[0] == name:
            return True
    return False



# Copyright 2019 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import traceback

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_grouped_by_priority

logger = colcon_logger.getChild(__name__)


class PrefixPathExtensionPoint:
    """
    The interface for prefix path extensions.

    A prefix path extension populates a list of chained prefix paths.

    For each instance the attribute `PREFIX_PATH_NAME` is being set to the
    basename of the entry point registering the extension.
    """

    """The version of the prefix path extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of prefix path extensions."""
    PRIORITY = 100

    def extend_prefix_path(self, paths):
        """
        Extend the given list of prefix paths.

        The items are ordered from higher to lower priority paths.

        This method must be overridden in a subclass.

        :param paths: The list of prefix paths to be extended
        :type paths: list
        """
        raise NotImplementedError()


def get_prefix_path_extensions():
    """
    Get the available prefix path extensions.

    The extensions are grouped by their priority and each group is ordered by
    the entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PREFIX_PATH_NAME = name
    return order_extensions_grouped_by_priority(extensions)


def get_chained_prefix_path(*, skip=None):
    """
    Get the chained prefix paths.

    The items are ordered from higher to lower priority paths.
    Repeated paths are skipped.

    :param skip: The current prefix path to be skipped and not be included in
      the return value
    :returns: The list of prefix paths
    :rtype: list
    """
    chained_prefix_path = []
    extensions = get_prefix_path_extensions()
    for priority in extensions.keys():
        extensions_same_prio = extensions[priority]
        for extension in extensions_same_prio.values():
            try:
                retval = extension.extend_prefix_path(chained_prefix_path)
                assert retval is None, \
                    'extend_prefix_path() should return None'
            except Exception as e:  # noqa: F841
                # catch exceptions raised in prefix path extension
                exc = traceback.format_exc()
                logger.error(
                    'Exception in prefix path extension '
                    "'{extension.PREFIX_PATH_NAME}': {e}\n{exc}"
                    .format_map(locals()))
                # skip failing extension, continue with next one
    unique_prefix_path = []
    for p in chained_prefix_path:
        if skip is not None and str(p) == str(skip):
            continue
        if p in unique_prefix_path:
            continue
        unique_prefix_path.append(p)
    return unique_prefix_path



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
from pathlib import PurePosixPath
import sys

from colcon_core.event.test import TestFailure
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from colcon_core.task import check_call
from colcon_core.task.python.test import has_test_dependency
from colcon_core.task.python.test import PythonTestingStepExtensionPoint
from colcon_core.verb.test import logger
from pkg_resources import iter_entry_points
from pkg_resources import parse_version


class PytestPythonTestingStep(PythonTestingStepExtensionPoint):
    """Use `pytest` to test Python packages."""

    # use a higher priority than the default priority
    # in order to become the default
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PythonTestingStepExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        entry_points = iter_entry_points('distutils.commands', name='pytest')
        if not list(entry_points):
            raise SkipExtensionException("'pytest-runner' not found")

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--pytest-args',
            nargs='*', metavar='*', type=str.lstrip,
            help='Pass arguments to pytests. '
            'Arguments matching other options must be prefixed by a space,\n'
            'e.g. --pytest-args " --help"')
        parser.add_argument(
            '--pytest-with-coverage',
            action='store_true',
            help='Generate coverage information')

    def match(self, context, env, setup_py_data):  # noqa: D102
        return has_test_dependency(setup_py_data, 'pytest')

    async def step(self, context, env, setup_py_data):  # noqa: D102
        cmd = [
            sys.executable,
            'setup.py', 'pytest',
            'egg_info', '--egg-base', context.args.build_base,
        ]
        junit_xml_path = Path(
            context.args.test_result_base
            if context.args.test_result_base
            else context.args.build_base) / 'pytest.xml'
        # avoid using backslashes in the PYTEST_ADDOPTS env var on Windows
        args = [
            '--tb=short',
            '--junit-xml=' + str(PurePosixPath(*junit_xml_path.parts)),
            '--junit-prefix=' + context.pkg.name,
        ]
        # use -o option only when available
        # https://github.com/pytest-dev/pytest/blob/3.3.0/CHANGELOG.rst
        from pytest import __version__ as pytest_version
        if parse_version(pytest_version) >= parse_version('3.3.0'):
            args += [
                '-o', 'cache_dir=' + str(PurePosixPath(
                    *(Path(context.args.build_base).parts)) / '.pytest_cache'),
            ]
        env = dict(env)

        if (
            context.args.pytest_with_coverage or
            has_test_dependency(setup_py_data, 'pytest-cov')
        ):
            try:
                from pytest_cov import __version__ as pytest_cov_version
            except ImportError:
                logger.warning(
                    'Test coverage will not be produced for package '
                    "'{context.pkg.name}' since the pytest extension 'cov' "
                    'was not found'.format_map(locals()))
            else:
                args += [
                    '--cov=' + str(PurePosixPath(
                        *(Path(context.args.path).parts))),
                    '--cov-report=html:' + str(PurePosixPath(
                        *(Path(context.args.build_base).parts)) /
                        'coverage.html'),
                    '--cov-report=xml:' + str(PurePosixPath(
                        *(Path(context.args.build_base).parts)) /
                        'coverage.xml'),
                ]
                # use --cov-branch option only when available
                # https://github.com/pytest-dev/pytest-cov/blob/v2.5.0/CHANGELOG.rst
                if parse_version(pytest_cov_version) >= parse_version('2.5.0'):
                    args += [
                        '--cov-branch',
                    ]
                else:
                    logger.warning(
                        'Test coverage will be produced but will not contain '
                        'branch coverage information because the pytest '
                        "extension 'cov' does not support it (need 2.5.0, "
                        'have {pytest_cov_version})'.format_map(locals()))
                env['COVERAGE_FILE'] = os.path.join(
                    context.args.build_base, '.coverage')

        if context.args.retest_until_fail:
            try:
                import pytest_repeat  # noqa: F401
            except ImportError:
                logger.warning(
                    "Ignored '--retest-until-fail' for package "
                    "'{context.pkg.name}' since the pytest extension 'repeat' "
                    'was not found'.format_map(locals()))
            else:
                count = context.args.retest_until_fail + 1
                args += [
                    '--count={count}'.format_map(locals()),
                ]

        if context.args.retest_until_pass:
            try:
                import pytest_rerunfailures  # noqa: F401
            except ImportError:
                logger.warning(
                    "Ignored '--retest-until-pass' for package "
                    "'{context.pkg.name}' since pytest extension "
                    "'rerunfailures' was not found".format_map(locals()))
            else:
                args += [
                    '--reruns={context.args.retest_until_pass}'
                    .format_map(locals()),
                ]

        if context.args.pytest_args is not None:
            args += context.args.pytest_args

        if args:
            env['PYTEST_ADDOPTS'] = ' '.join(
                a if ' ' not in a else '"{a}"'.format_map(locals())
                for a in args)

        # create dummy result in case the invocation fails early
        # and doesn't generate a result file at all
        junit_xml_path.parent.mkdir(parents=True, exist_ok=True)
        junit_xml_path.write_text("""<?xml version="1.0" encoding="UTF-8"?>
<testsuite name="{context.pkg.name}" tests="1" failures="1" time="0" errors="0" skip="0">
  <testcase classname="{context.pkg.name}" name="pytest.missing_result" status="run" time="0">
    <failure message="The test invocation failed without generating a result file."/>
  </testcase>
</testsuite>
""".format_map(locals()))  # noqa: E501

        rc = await check_call(context, cmd, cwd=context.args.path, env=env)

        # use local import to avoid a dependency on pytest
        try:
            from _pytest.main import ExitCode
            EXIT_CODE_TESTS_FAILED = ExitCode.TESTS_FAILED  # noqa: N806
        except ImportError:
            # support pytest < 5.0
            from _pytest.main import EXIT_TESTSFAILED
            EXIT_CODE_TESTS_FAILED = EXIT_TESTSFAILED  # noqa: N806
        if rc and rc.returncode == EXIT_CODE_TESTS_FAILED:
            context.put_event_into_queue(
                TestFailure(context.pkg.name))

        try:
            from _pytest.main import ExitCode
            EXIT_CODE_NO_TESTS = ExitCode.NO_TESTS_COLLECTED  # noqa: N806
        except ImportError:
            # support pytest < 5.0
            from _pytest.main import EXIT_NOTESTSCOLLECTED
            EXIT_CODE_NO_TESTS = EXIT_NOTESTSCOLLECTED  # noqa: N806
        if rc and rc.returncode not in (
            EXIT_CODE_NO_TESTS, EXIT_CODE_TESTS_FAILED
        ):
            return rc.returncode



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import uuid

from colcon_core.logging import colcon_logger

logger = colcon_logger.getChild(__name__)

_config_path = None
_config_path_env_var = None
_log_base_path = None
_log_base_path_default = None
_log_base_path_env_var = None
_log_subdirectory = None


def get_config_path():
    """
    Get the base path for configuration files.

    :function:`set_default_config_path` must have been called before.

    :returns: The base path for configuration files
    :rtype: Path
    """
    global _config_path_env_var
    if _config_path_env_var is not None:
        path = os.environ.get(_config_path_env_var)
        if path:
            return Path(str(path))
    global _config_path
    assert _config_path is not None
    return _config_path


def set_default_config_path(*, path, env_var=None):
    """
    Set the base path for configuration files.

    Optionally an environment variable name can be provided which if set will
    override the configured base path.

    An info message is logged which states the used path.

    :param path: The base path
    :param str env_var: The name of the environment variable
    """
    global _config_path
    global _config_path_env_var
    _config_path = Path(str(path))
    _config_path_env_var = env_var
    config_path = get_config_path()
    logger.info("Using config path '{config_path}'".format_map(locals()))


def get_log_path():
    """
    Get the base path for logging.

    :function:`set_default_log_path` must have been called before.

    :returns: The base path for logging or None if logging is disabled
    :rtype: Path or None
    """
    global _log_base_path
    global _log_base_path_env_var
    path = None
    if _log_base_path is not None:
        path = _log_base_path
    elif (
        _log_base_path_env_var is not None and
        os.environ.get(_log_base_path_env_var)
    ):
        path = os.environ.get(_log_base_path_env_var)
    else:
        global _log_base_path_default
        path = _log_base_path_default

    if path == os.devnull:
        return None

    return Path(str(path)) / _log_subdirectory


def set_default_log_path(
    *, base_path, env_var=None, subdirectory=None, default='log'
):
    """
    Set the base path for logging.

    Optionally an environment variable name can be provided which if set will
    override the configured base path.

    An info message is logged which states the used path.

    :param base_path: The base path
    :param str env_var: The name of the environment variable
    :param str subdirectory: The name of the subdirectory, if not provided a
      random uuid will be used instead
    :param default: The default base path if the passed base path is None and
      the environment variable is not set
    """
    global _log_base_path
    global _log_base_path_default
    global _log_base_path_env_var
    global _log_subdirectory
    _log_base_path = base_path
    _log_base_path_default = default
    _log_base_path_env_var = env_var
    assert subdirectory is None or subdirectory
    _log_subdirectory = subdirectory \
        if subdirectory is not None \
        else str(uuid.uuid4())


_create_log_path_called = False


def create_log_path(verb_name):
    """
    Create a not yet existing logging directory.

    The logging directory returned by :function:`get_log_path` must not yet
    exist on the first call of this function.
    If it does exist the function will append a serial number to the path until
    the path doesn't exist and can be created.

    Subsequent invocations of this function are noops.

    A `COLCON_IGNORE` marker file is being placed in the parent directory of
    the logging directory to avoid it being crawled for packages.

    Two symlinks are created as siblings of the log path:
    * `latest_<verb_name>` linking to the log path
    * `latest` linking to `latest_<verb_name>`

    On Windows platforms Administrator privileges are required to create
    these symlinks.
    Otherwise they are being skipped.

    :param str verb_name: The verb name
    """
    global _create_log_path_called
    if _create_log_path_called:
        return
    _create_log_path_called = True

    path = get_log_path()
    try:
        # try to create the directory
        os.makedirs(str(path))
    except FileExistsError:
        # if it already exists try again with serial number suffixes
        global _log_subdirectory
        assert path.name == _log_subdirectory
        suffix = 2
        while True:
            path_with_suffix = path.with_name(path.name + '_' + str(suffix))
            try:
                os.makedirs(str(path_with_suffix))
            except FileExistsError:
                suffix += 1
                assert suffix < 1000, 'Prevent infinite loop'
                continue
            _log_subdirectory = path_with_suffix.name
            assert get_log_path() == path_with_suffix
            path = path_with_suffix
            break

    logger.info("Using log path '{path}'".format_map(locals()))

    # ensure the base log path has an ignore marker file
    # to avoid recursively crawling through log directories
    from colcon_core.package_identification.ignore import IGNORE_MARKER
    ignore_marker = path.parent / IGNORE_MARKER
    ignore_marker.touch()

    # create latest symlinks
    _create_symlink(
        path, path.parent / 'latest_{verb_name}'.format_map(locals()))
    _create_symlink(
        path.parent / 'latest_{verb_name}'.format_map(locals()),
        path.parent / 'latest')


def _create_symlink(src, dst):
    if dst.exists():
        # directory exists or valid symlink
        if not dst.is_symlink():
            # do not change non symlink paths
            return
        if dst.resolve() == src.resolve():
            # desired symlink already exists
            return
    # remove valid symlink to wrong destination (previous if, no return)
    # or invalid symlink (non-existing else from previous if)
    if dst.is_symlink():
        try:
            dst.unlink()
        except FileNotFoundError:
            pass

    # use relative path when possible
    try:
        src = src.relative_to(dst.parent)
    except ValueError:
        pass
    try:
        os.symlink(str(src), str(dst))
    except FileNotFoundError:
        pass
    except OSError:
        # Administrator privileges are required on Windows
        pass


def get_relative_package_index_path():
    """
    Get the prefix-relative path to the package index.

    :returns: The relative path to the package index
    :rtype: Path
    """
    # the value is also being hard coded in shell/template/prefix_util.py
    return Path('share', 'colcon-core', 'packages')



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import traceback

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority

logger = colcon_logger.getChild(__name__)


class PackageDiscoveryExtensionPoint:
    """
    The interface for discovery extensions.

    A discovery extension provides potential locations of packages which are
    then being check by the identification extensions.

    For each instance the attribute `PACKAGE_DISCOVERY_NAME` is being set to
    the basename of the entry point registering the extension.
    """

    """The version of the discovery extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of discovery extensions."""
    PRIORITY = 100

    def has_default(self):
        """
        Check if the extension has a default parameter is none are provided.

        The method is intended to be overridden in a subclass.

        :param args: The parsed command line arguments
        :returns: True if `discover()` should be called even if no parameters
          are provided, False otherwise
        :rtype: bool
        """
        return False

    def add_arguments(self, *, parser, with_default):
        """
        Add command line arguments specific to the package discovery.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        :param bool with_default: if True the extension should add a default
          value to at least one of the added arguments, otherwise not
        """
        pass

    def has_parameters(self, *, args):
        """
        Check if parameters have been passed for this extension.

        This method must be overridden in a subclass.

        :param args: The parsed command line arguments
        :returns: True if `discover()` should be called, False otherwise
        :rtype: bool
        """
        raise NotImplementedError()

    def discover(self, *, args, identification_extensions):
        """
        Discover packages using the passed identification extensions.

        The method is intended to be overridden in a subclass.
        If the `has_parameters` method never returns True this method is never
        invoked and therefore doesn't have to be implemented.

        :param args: The parsed command line arguments
        :param identification_extensions: The identification extensions
        :returns: set of
          :py:class:`colcon_core.package_descriptor.PackageDescriptor`
        :rtype: set
        """
        raise NotImplementedError()


def get_package_discovery_extensions():
    """
    Get the available package discovery extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_DISCOVERY_NAME = name
    return order_extensions_by_priority(extensions)


def add_package_discovery_arguments(parser, *, extensions=None):
    """
    Add the command line arguments for the package discovery extensions.

    :param parser: The argument parser
    :param extensions: The package discovery extensions to use, if `None` is
      passed use the extensions provided by
      :function:`get_package_discovery_extensions`
    """
    if extensions is None:
        extensions = get_package_discovery_extensions()
    group = parser.add_argument_group(title='Discovery arguments')

    # find the first extension which has default values
    first_extension_with_default = None
    for name, extension in extensions.items():
        try:
            has_default = extension.has_default()
        except Exception as e:  # noqa: F841
            # catch exceptions raised in discovery extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package discovery extension '
                "'{extension.PACKAGE_DISCOVERY_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one
        else:
            if has_default:
                first_extension_with_default = name
                break

    # collect arguments from the extensions
    for name, extension in extensions.items():
        with_default = name == first_extension_with_default
        try:
            retval = extension.add_arguments(
                parser=group, with_default=with_default)
            assert retval is None, 'add_arguments() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in discovery extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package discovery extension '
                "'{extension.PACKAGE_DISCOVERY_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one


def discover_packages(
    args, identification_extensions, *, discovery_extensions=None
):
    """
    Discover and identify packages.

    All discovery extensions which report to have parameters are being used to
    discover packages.
    If none report to have parameters all discovery extensions are being used
    but only the one with a default value should discover packages.

    Each discovery extension uses the passed identification extensions to check
    each potential location for the existence of a package.

    :param args: The parsed command line arguments
    :param identification_extensions: The package identification extensions to
      pass to each invocation of
      :function:`PackageDiscoveryExtensionPoint.discover`
    :param discovery_extensions: The package discovery extensions to use, if
      `None` is passed use the extensions provided by
      :function:`get_package_discovery_extensions`
    :returns: set of
      :py:class:`colcon_core.package_descriptor.PackageDescriptor`
    :rtype: set
    """
    if discovery_extensions is None:
        discovery_extensions = get_package_discovery_extensions()
    if not discovery_extensions:
        logger.warning('No package discovery extensions found')
        return set()

    # use only the discovery extensions which have parameters if any
    # otherwise use all
    discovery_extensions_with_parameters = _get_extensions_with_parameters(
        args, discovery_extensions)
    if discovery_extensions_with_parameters:
        discovery_extensions = discovery_extensions_with_parameters

    return _discover_packages(
        args, identification_extensions, discovery_extensions)


def _get_extensions_with_parameters(
    args, discovery_extensions
):
    with_parameters = OrderedDict()
    for extension in discovery_extensions.values():
        logger.log(
            1,
            'discover_packages({extension.PACKAGE_DISCOVERY_NAME}) check '
            'parameters'.format_map(locals()))
        try:
            has_parameter = extension.has_parameters(args=args)
        except Exception as e:  # noqa: F841
            # catch exceptions raised in discovery extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package discovery extension '
                "'{extension.PACKAGE_DISCOVERY_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one
        else:
            if has_parameter:
                with_parameters[extension.PACKAGE_DISCOVERY_NAME] = extension
    return with_parameters


def _discover_packages(
    args, identification_extensions, discovery_extensions
):
    all_descs = set()
    # if none had explicit parameters use the first which has defaults
    for extension in discovery_extensions.values():
        logger.log(
            1, 'discover_packages(%s) discover',
            extension.PACKAGE_DISCOVERY_NAME)
        try:
            descs = extension.discover(
                args=args, identification_extensions=identification_extensions)
            assert isinstance(descs, set), 'discover() should return a set'
        except NotImplementedError:
            # skip extension not implementing discovery
            continue
        except Exception as e:  # noqa: F841
            # catch exceptions raised in discovery extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package discovery extension '
                "'{extension.PACKAGE_DISCOVERY_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one
            continue
        else:
            logger.log(
                1, 'discover_packages(%s) using defaults',
                extension.PACKAGE_DISCOVERY_NAME)
            all_descs |= descs
    return all_descs



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import sys
from tempfile import TemporaryDirectory

from colcon_core.event.command import Command
from colcon_core.event.job import JobProgress
from colcon_core.event.output import StderrLine
from colcon_core.event.output import StdoutLine
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.task import add_task_arguments
from colcon_core.task import check_call
from colcon_core.task import create_file
from colcon_core.task import get_task_extension
from colcon_core.task import get_task_extensions
from colcon_core.task import install
from colcon_core.task import TaskContext
from colcon_core.task import TaskExtensionPoint
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext
from .run_until_complete import run_until_complete


def test_context_interface():
    context = TaskContext(pkg=None, args=None, dependencies=None)
    with pytest.raises(NotImplementedError):
        context.put_event_into_queue(None)


class Extension(TaskExtensionPoint):
    TASK_NAME = 'do'

    async def do(self, *args, **kwargs):
        self.progress('progress')
        self.print('hello')
        self.print('hello', file=sys.stdout)
        self.print('world', file=sys.stderr)
        with pytest.raises(AssertionError):
            self.print('invalid file handle', file=False)
        return 1


def test_extension_interface():
    context = Mock()

    # capture events
    events = []

    def put_event_into_queue(event):
        nonlocal events
        events.append(event)

    context.put_event_into_queue = put_event_into_queue

    extension = Extension()
    extension.set_context(context=context)
    rc = run_until_complete(extension())
    assert rc == 1

    assert len(events) == 4
    assert isinstance(events[0], JobProgress)
    assert events[0].progress == 'progress'
    assert isinstance(events[1], StdoutLine)
    assert events[1].line == 'hello\n'
    assert isinstance(events[2], StdoutLine)
    assert events[2].line == 'hello\n'
    assert isinstance(events[3], StderrLine)
    assert events[3].line == 'world\n'


# TODO figure out how to avoid the stderr output
@pytest.mark.skip(
    reason='Results in stderr output due to a UnicodeDecodeError for the '
           'generated coverage files')
def test_check_call():
    context = Mock()
    events = []

    def put_event_into_queue(event):
        nonlocal events
        events.append(event)

    context.put_event_into_queue = put_event_into_queue
    cmd = [
        sys.executable, '-c',
        "import sys; print('hello'); print('world', file=sys.stderr)"]
    coroutine = check_call(context, cmd)
    completed_process = run_until_complete(coroutine)
    assert completed_process.returncode == 0
    assert len(events) == 3
    assert isinstance(events[0], Command)
    assert events[0].cmd == cmd
    assert isinstance(events[1], StdoutLine)
    assert events[1].line == b'hello\n'
    assert isinstance(events[2], StderrLine)
    assert events[2].line == b'world\n'


class Extension1(TaskExtensionPoint):

    def build(self, *args, **kwargs):
        pass  # pragma: no cover


class Extension2(TaskExtensionPoint):

    def build(self, *args, **kwargs):
        pass  # pragma: no cover


def instantiate_extensions_without_cache(
    group_name, *, exclude_names=None, unique_instance=False
):
    return instantiate_extensions(group_name)


def test_add_task_arguments():
    parser = Mock()
    task_name = 'colcon_core.task.build'
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        with patch(
            'colcon_core.task.instantiate_extensions',
            side_effect=instantiate_extensions_without_cache
        ):
            extensions = get_task_extensions(task_name)
            # one exception, one success
            extensions['extension1'].add_arguments = Mock(
                side_effect=RuntimeError('custom exception'))
            with patch('colcon_core.task.logger.error') as error:
                add_task_arguments(parser, task_name)
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in task extension 'build.extension1': custom "
                'exception\n')

            # invalid return value
            extensions['extension1'].add_arguments = Mock()
            extensions['extension2'].add_arguments = Mock(return_value=None)
            with patch('colcon_core.task.logger.error') as error:
                add_task_arguments(parser, task_name)
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in task extension 'build.extension1': "
                'add_arguments() should return None\n')
            assert extensions['extension2'].add_arguments.call_count == 1


def test_get_task_extension():
    task_name = 'colcon_core.task.build'
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        # request invalid extension
        extension = get_task_extension(task_name, 'package_type')
        assert extension is None

        # request valid extension
        extension = get_task_extension(task_name, 'extension2')
        assert isinstance(extension, Extension2)


def test_create_file():
    with TemporaryDirectory(prefix='test_colcon_') as base_path:
        args = Mock()
        args.install_base = base_path

        create_file(args, 'file.txt')
        path = Path(base_path) / 'file.txt'
        assert path.is_file()
        assert path.read_text() == ''

        create_file(args, 'path/file.txt', content='content')
        path = Path(base_path) / 'path' / 'file.txt'
        assert path.is_file()
        assert path.read_text() == 'content'


def test_install():
    with TemporaryDirectory(prefix='test_colcon_') as base_path:
        args = Mock()
        args.path = os.path.join(base_path, 'path')
        args.install_base = os.path.join(base_path, 'install')
        args.symlink_install = False

        # create source files
        os.makedirs(args.path)
        with open(os.path.join(args.path, 'source.txt'), 'w') as h:
            h.write('content')
        with open(os.path.join(args.path, 'source2.txt'), 'w') as h:
            h.write('content2')

        # copy file
        install(args, 'source.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert not path.is_symlink()
        assert path.read_text() == 'content'

        # skip all symlink tests on Windows for now
        if sys.platform == 'win32':  # pragma: no cover
            return

        # symlink file, removing existing file
        args.symlink_install = True
        install(args, 'source.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert path.is_symlink()
        assert path.samefile(os.path.join(args.path, 'source.txt'))
        assert path.read_text() == 'content'

        # symlink other file, removing existing directory
        os.remove(os.path.join(args.install_base, 'destination.txt'))
        os.makedirs(os.path.join(args.install_base, 'destination.txt'))
        install(args, 'source2.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert path.is_symlink()
        assert path.samefile(os.path.join(args.path, 'source2.txt'))
        assert path.read_text() == 'content2'

        # copy file, removing existing symlink
        args.symlink_install = False
        install(args, 'source.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert not path.is_symlink()
        assert path.read_text() == 'content'

        # symlink file
        os.remove(os.path.join(args.install_base, 'destination.txt'))
        args.symlink_install = True
        install(args, 'source.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert path.is_symlink()
        assert path.samefile(os.path.join(args.path, 'source.txt'))
        assert path.read_text() == 'content'

        # symlink file, same already existing
        install(args, 'source.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert path.is_symlink()
        assert path.samefile(os.path.join(args.path, 'source.txt'))
        assert path.read_text() == 'content'

        # symlink exists, but to a not existing location
        os.remove(os.path.join(args.path, 'source.txt'))
        install(args, 'source2.txt', 'destination.txt')
        path = Path(base_path) / 'install' / 'destination.txt'
        assert path.is_file()
        assert path.is_symlink()
        assert path.samefile(os.path.join(args.path, 'source2.txt'))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import sys

from colcon_core.event.command import Command

from .environment_context import EnvironmentContext


def test_command_to_string():
    cmd = ['executable', 'arg1', 'arg2']
    cwd = '/some/path'
    command = Command(cmd, cwd=cwd)
    assert command.to_string() == \
        "Invoking command in '/some/path': executable arg1 arg2"

    env = {
        '_TEST_NEW_KEY': 'new',
        '_TEST_SAME_VALUE': 'same',
        '_TEST_DIFFERENT_VALUE': 'different',
        '_TEST_PREPENDED_VALUE': 'before-base',
        '_TEST_APPENDED_VALUE': 'base-after',
    }
    if sys.platform != 'win32':
        env['PWD'] = '/other/path'
    command = Command(cmd, cwd=cwd, env=env)

    expected = "Invoking command in '/some/path': " \
        '_TEST_APPENDED_VALUE=${_TEST_APPENDED_VALUE}-after ' \
        '_TEST_DIFFERENT_VALUE=different ' \
        '_TEST_NEW_KEY=new ' \
        '_TEST_PREPENDED_VALUE=before-${_TEST_PREPENDED_VALUE} ' \
        'executable arg1 arg2'
    if sys.platform == 'win32':
        expected = expected.replace('${', '%')
        expected = expected.replace('}', '%')
    with EnvironmentContext(
        _TEST_SAME_VALUE='same',
        _TEST_DIFFERENT_VALUE='same',
        _TEST_PREPENDED_VALUE='base',
        _TEST_APPENDED_VALUE='base',
    ):
        assert command.to_string() == expected

    cmd = ['executable', '&&', 'other exec']
    command = Command(cmd, cwd=cwd)
    assert command.to_string() == \
        "Invoking command in '/some/path': executable && other exec"
    command = Command(cmd, cwd=cwd, shell=True)
    if sys.platform != 'win32':
        assert command.to_string() == \
            "Invoking command in '/some/path': executable && 'other exec'"
    else:
        assert command.to_string() == \
            "Invoking command in '/some/path': " \
            'executable && "other exec"'



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import logging

from colcon_core.logging import get_numeric_log_level
from colcon_core.logging import set_logger_level_from_env
from mock import Mock
import pytest

from .environment_context import EnvironmentContext


def test_set_logger_level_from_env():
    logger = logging.getLogger('test')
    default_level = logger.getEffectiveLevel()

    # not set
    set_logger_level_from_env(logger, 'COLCON_TEST_LOGGER_LEVEL')
    assert logger.getEffectiveLevel() == default_level

    # invalid value
    with EnvironmentContext(COLCON_TEST_LOGGER_LEVEL='invalid'):
        logger.warning = Mock()
        set_logger_level_from_env(logger, 'COLCON_TEST_LOGGER_LEVEL')
        assert logger.warning.call_count == 1
    assert logger.getEffectiveLevel() == default_level

    # valid value
    with EnvironmentContext(COLCON_TEST_LOGGER_LEVEL='debug'):
        set_logger_level_from_env(logger, 'COLCON_TEST_LOGGER_LEVEL')
    assert logger.getEffectiveLevel() == logging.DEBUG


def test_get_numeric_log_level():
    # numeric
    log_level = get_numeric_log_level('10')
    assert log_level == logging.DEBUG

    # string
    log_level = get_numeric_log_level('info')
    assert log_level == logging.INFO

    # string with mixed case
    log_level = get_numeric_log_level('Warning')
    assert log_level == logging.WARNING

    # invalid string
    with pytest.raises(ValueError) as e:
        get_numeric_log_level('invalid')
    assert str(e.value).endswith(
        'valid names are: CRITICAL, ERROR, WARNING, INFO, DEBUG '
        '(case-insensitive)')

    # negative numeric
    with pytest.raises(ValueError) as e:
        get_numeric_log_level('-1')
    assert str(e.value).endswith('numeric log levels must be positive')



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.shell.template import expand_template
from em import TransientParseError
from mock import patch
import pytest


def test_expand_template():
    with TemporaryDirectory(prefix='test_colcon_') as destination_path:
        template_path = Path(destination_path) / 'invalid_template.em'

        template_path.write_text(
            '@[if True]')
        with pytest.raises(TransientParseError):
            with patch('colcon_core.shell.template.logger.error') as error:
                expand_template(template_path, destination_path, {})
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].endswith(
            " processing template '{template_path}'".format_map(locals()))

        template_path.write_text(
            '@(var)')
        with pytest.raises(NameError):
            with patch('colcon_core.shell.template.logger.error') as error:
                expand_template(template_path, destination_path, {})
        # the raised exception is catched and results in an error message
        assert error.call_count == 1
        assert len(error.call_args[0]) == 1
        assert error.call_args[0][0].endswith(
            " processing template '{template_path}'".format_map(locals()))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import namedtuple


EnvironmentVariable = namedtuple(
    'EnvironmentVariable', ('name', 'description'))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import copy
import traceback
from typing import Dict
from typing import Union

from colcon_core.logging import colcon_logger
from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_grouped_by_priority

logger = colcon_logger.getChild(__name__)


class IgnoreLocationException(Exception):
    """
    Exception to signal that a path should be skipped.

    This also excludes all recursive subdirectories from being considered.

    It should be raised in the
    :function:`PackageIdentificationExtensionPoint.identify` method of package
    identification extensions.
    """

    pass


class PackageIdentificationExtensionPoint:
    """
    The interface for package identification extensions.

    A package identification extension populates a package descriptor if a
    given location contains a package.

    For each instance the attribute `PACKAGE_IDENTIFICATION_NAME` is being set
    to the basename of the entry point registering the extension.
    """

    """The version of the package identification extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of package identification extensions."""
    PRIORITY = 100

    def identify(self, desc: PackageDescriptor):
        """
        Check if the given path contains a package.

        If yes, update the package descriptor with additional information.
        If the descriptor already has a package type only act if the type
        matches this extension.

        This method must be overridden in a subclass.

        :param desc: The package descriptor with the directory to check
        :type desc:
          :py:class:`colcon_core.package_descriptor.PackageDescriptor`
        :raises IgnoreLocationException: Skip the path as well as all recursive
          subdirectories
        """
        raise NotImplementedError()


def get_package_identification_extensions():
    """
    Get the available package identification extensions.

    The extensions are grouped by their priority and each group is ordered by
    the entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.PACKAGE_IDENTIFICATION_NAME = name
    return order_extensions_grouped_by_priority(extensions)


def identify(
    extensions: Dict[  # actually an OrderedDict
        int,  # priority
        Dict[str, PackageIdentificationExtensionPoint],  # an OrderedDict
    ],
    path: str,
) -> Union[None, bool, PackageDescriptor]:
    """
    Identify the package in the given path.

    :param extensions: dict of extensions
    :param path: The path
    """
    desc = PackageDescriptor(path)

    for extensions_same_prio in extensions.values():
        result = _identify(extensions_same_prio, desc)

        # continue with next priority level if no information was contributed
        if result is None:
            continue

        # skip location since identification is ambiguous
        if result is False:
            raise IgnoreLocationException()

        assert isinstance(result, PackageDescriptor), result
        if result.identifies_package():
            return result

        # use incrementally populated descriptor for next priority level
        desc = result

    if getattr(desc, 'type', None) or getattr(desc, 'name', None):
        logger.warning(
            "package '{desc.path}' has type or name but is incomplete"
            .format_map(locals()))
    return None


# the following variable only exists to avoid repeatedly copying descriptors
_reused_descriptor_instance = None


def _identify(extensions_same_prio, desc):
    global _reused_descriptor_instance
    logger.log(
        1, '_identify(%s) by extensions %s',
        desc.path, sorted(extensions_same_prio.keys()))
    # collect the optionally modified descriptors from all extensions
    results = {desc}
    for key, extension in extensions_same_prio.items():
        # create copy of the descriptor if the reusable instance is different
        if (
            _reused_descriptor_instance is None or
            not _are_descriptors_equal(desc, _reused_descriptor_instance)
        ):
            _reused_descriptor_instance = copy.deepcopy(desc)

        logger.log(1, "_identify(%s) by extension '%s'", desc.path, key)
        try:
            retval = extension.identify(_reused_descriptor_instance)
            assert retval is None, 'identify() should return None'
        except IgnoreLocationException:
            logger.log(1, '_identify(%s) ignored', desc.path)
            raise
        except Exception as e:  # noqa: F841
            # catch exceptions raised in identification extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in package identification extension '
                "'{extension.PACKAGE_IDENTIFICATION_NAME}' in '{desc.path}': "
                '{e}\n{exc}'.format_map(locals()))
            # skip failing extension, continue with next one
            continue

        # only add the descriptor if it is different from the existing result
        # if it is equal it can be attempted to reuse the instance instead
        if not _are_descriptors_equal(desc, _reused_descriptor_instance):
            results.add(_reused_descriptor_instance)
            # a new copy of the descriptor needs to be created next cycle
            _reused_descriptor_instance = None

    # multiple extensions populated the descriptor with different values
    if len(results) > 2:
        logger.warning(
            '_identify({desc.path}) has multiple matches and therefore is '
            'being ignored: '.format_map(locals()) +
            ', '.join(sorted(d.type for d in results if d.type is not None)))
        return False

    # no extension populated the descriptor with any values
    if len(results) == 1:
        return None

    results.remove(desc)
    desc = results.pop()
    logger.debug(
        "Package '{desc.path}' with type '{desc.type}' and name '{desc.name}'"
        .format_map(locals()))
    return desc


def _are_descriptors_equal(desc1, desc2):
    for s in PackageDescriptor.__slots__:
        if getattr(desc1, s) != getattr(desc2, s):
            return False
    return True



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from queue import Queue
import time

from colcon_core.event.timer import TimerEvent
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_reactor import create_event_reactor
from colcon_core.event_reactor import EventReactorShutdown
from mock import Mock
from mock import patch

from .entry_point_context import EntryPointContext


class CustomExtension(EventHandlerExtensionPoint):

    def __init__(self):
        super().__init__()
        self.events = []

    def __call__(self, event):
        self.events.append(event)


class Extension1(CustomExtension):
    pass


class Extension2(CustomExtension):

    def __init__(self):
        super().__init__()
        self.enabled = False


class Extension3(CustomExtension):

    def __call__(self, event):
        super().__call__(event)
        if event[0] in ('first', 'third'):
            raise RuntimeError('custom exception')


def test_create_event_reactor():
    context = Mock()
    context.args = Mock()
    context.args.event_handlers = []
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2, extension3=Extension3
    ):
        event_reactor = create_event_reactor(context)

    with event_reactor:
        queue = event_reactor.get_queue()
        assert isinstance(queue, Queue)

        # use larger interval to prevent different timing to effect the results
        event_reactor.TIMER_INTERVAL = 0.25

        # add a few dummy events
        with patch('colcon_core.event_reactor.logger.error') as error:
            queue.put(('first', None))
            queue.put(('second', None))
            queue.put(('third', None))

            # check the collected events so far
            event_reactor.flush()
        assert len(event_reactor._observers[0].events) == 4
        assert len(event_reactor._observers[1].events) == 4
        assert isinstance(event_reactor._observers[0].events[0][0], TimerEvent)
        assert isinstance(event_reactor._observers[1].events[0][0], TimerEvent)
        assert event_reactor._observers[0].events[1:] == \
            [('first', None), ('second', None), ('third', None)]
        assert event_reactor._observers[1].events[1:] == \
            [('first', None), ('second', None), ('third', None)]

        # the raised exception is catched and results in an error message
        assert error.call_count == 2
        assert len(error.call_args_list[0][0]) == 1
        assert error.call_args_list[0][0][0].startswith(
            "Exception in event handler extension 'extension3': "
            'custom exception\n')
        assert len(error.call_args_list[1][0]) == 1
        assert error.call_args_list[1][0][0].startswith(
            "Exception in event handler extension 'extension3': "
            'custom exception\n')

        # wait for another timer event to be generated
        time.sleep(event_reactor.TIMER_INTERVAL)
        assert len(event_reactor._observers[0].events) == 5
        assert len(event_reactor._observers[1].events) == 5
        assert isinstance(
            event_reactor._observers[0].events[-1][0], TimerEvent)
        assert isinstance(
            event_reactor._observers[1].events[-1][0], TimerEvent)

    assert len(event_reactor._observers[0].events) == 6
    assert len(event_reactor._observers[1].events) == 6
    assert isinstance(
        event_reactor._observers[0].events[-1][0], EventReactorShutdown)
    assert isinstance(
        event_reactor._observers[1].events[-1][0], EventReactorShutdown)

    # no harm in flushing after the thread has been joined
    event_reactor.flush()



# Copyright 2019 Rover Robotics
# Licensed under the Apache License, Version 2.0

import asyncio
from pathlib import Path
from tempfile import TemporaryDirectory
from types import SimpleNamespace

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.plugin_system import SkipExtensionException
import colcon_core.shell
from colcon_core.shell.bat import BatShell
from colcon_core.shell.sh import ShShell
from colcon_core.subprocess import new_event_loop
from colcon_core.task import TaskContext
from colcon_core.task.python.build import PythonBuildTask
import pytest


@pytest.fixture(autouse=True)
def monkey_patch_get_shell_extensions(monkeypatch):
    a_shell = None
    for shell_extension_class in [ShShell, BatShell]:
        try:
            a_shell = shell_extension_class()
            break
        except SkipExtensionException:
            pass

    if a_shell is None:
        pytest.fail('No valid shell extension found.')

    monkeypatch.setattr(
        colcon_core.shell,
        'get_shell_extensions',
        lambda: {
            200: {'mock': a_shell}
        }
    )


@pytest.fixture(autouse=True)
def monkey_patch_put_event_into_queue(monkeypatch):
    monkeypatch.setattr(
        TaskContext, 'put_event_into_queue', lambda *args: None
    )


def test_build_package():
    event_loop = new_event_loop()
    asyncio.set_event_loop(event_loop)
    try:
        with TemporaryDirectory(prefix='test_colcon_') as tmp_path_str:
            tmp_path = Path(tmp_path_str)
            python_build_task = PythonBuildTask()
            package = PackageDescriptor(tmp_path / 'src')
            package.name = 'test_package'
            package.type = 'python'

            context = TaskContext(
                pkg=package,
                args=SimpleNamespace(
                    path=str(tmp_path / 'src'),
                    build_base=str(tmp_path / 'build'),
                    install_base=str(tmp_path / 'install'),
                    symlink_install=False,
                ),
                dependencies={}
            )
            python_build_task.set_context(context=context)

            pkg = python_build_task.context.pkg

            pkg.path.mkdir()
            (pkg.path / 'setup.py').write_text(
                'from setuptools import setup\n'
                'setup(\n'
                '    name="test_package",\n'
                '    packages=["my_module"],\n'
                ')\n'
            )
            (pkg.path / 'my_module').mkdir()
            (pkg.path / 'my_module' / '__init__.py').touch()

            src_base = Path(python_build_task.context.args.path)

            source_files_before = set(src_base.rglob('*'))
            event_loop.run_until_complete(python_build_task.build())
            source_files_after = set(src_base.rglob('*'))
            assert source_files_before == source_files_after

            build_base = Path(python_build_task.context.args.build_base)
            assert 1 == len(list(build_base.rglob('my_module/__init__.py')))

            install_base = Path(python_build_task.context.args.install_base)
            assert 1 == len(list(install_base.rglob('my_module/__init__.py')))

            pkg_info, = install_base.rglob('PKG-INFO')
            assert 'Name: test-package' in pkg_info.read_text().splitlines()
    finally:
        event_loop.close()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from collections import OrderedDict
import os
import types

from colcon_core.argument_parser.destination_collector \
    import DestinationCollectorDecorator
from colcon_core.event.test import TestFailure
from colcon_core.event_handler import add_event_handler_arguments
from colcon_core.executor import add_executor_arguments
from colcon_core.executor import execute_jobs
from colcon_core.executor import Job
from colcon_core.executor import OnError
from colcon_core.logging import colcon_logger
from colcon_core.package_selection import add_arguments \
    as add_packages_arguments
from colcon_core.package_selection import get_packages
from colcon_core.plugin_system import satisfies_version
from colcon_core.task import add_task_arguments
from colcon_core.task import get_task_extension
from colcon_core.task import TaskContext
from colcon_core.verb import check_and_mark_build_tool
from colcon_core.verb import check_and_mark_install_layout
from colcon_core.verb import update_object
from colcon_core.verb import VerbExtensionPoint

logger = colcon_logger.getChild(__name__)


class TestPackageArguments:
    """Arguments to test a specific package."""

    def __init__(self, pkg, args, *, additional_destinations=None):
        """
        Construct a TestPackageArguments.

        :param pkg: The package descriptor
        :param args: The parsed command line arguments
        :param list additional_destinations: The destinations of additional
          arguments
        """
        super().__init__()
        self.path = os.path.abspath(
            os.path.join(os.getcwd(), str(pkg.path)))
        self.build_base = os.path.abspath(os.path.join(
            os.getcwd(), args.build_base, pkg.name))
        self.install_base = os.path.abspath(os.path.join(
            os.getcwd(), args.install_base))
        if not args.merge_install:
            self.install_base = os.path.join(
                self.install_base, pkg.name)
        self.test_result_base = os.path.abspath(os.path.join(
            os.getcwd(), args.test_result_base, pkg.name)) \
            if args.test_result_base else None

        # set additional arguments
        for dest in (additional_destinations or []):
            # from the command line
            if hasattr(args, dest):
                update_object(
                    self, dest, getattr(args, dest),
                    pkg.name, 'test', 'command line')
            # from the package metadata
            if dest in pkg.metadata:
                update_object(
                    self, dest, pkg.metadata[dest],
                    pkg.name, 'test', 'package metadata')


class TestVerb(VerbExtensionPoint):
    """
    Test a set of packages.

    Each test task is expected to post a
    :py:class:`colcon_core.event.test.TestFailure` event in case of test
    failures.
    """

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(VerbExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def add_arguments(self, *, parser):  # noqa: D102
        parser.add_argument(
            '--build-base',
            default='build',
            help='The base path for all build directories (default: build)')
        parser.add_argument(
            '--install-base',
            default='install',
            help='The base path for all install prefixes (default: install)')
        parser.add_argument(
            '--merge-install',
            action='store_true',
            help='Merge all install prefixes into a single location')
        parser.add_argument(
            '--test-result-base',
            help='The base path for all test results (default: --build-base)')
        group = parser.add_mutually_exclusive_group()
        group.add_argument(
            '--retest-until-fail',
            type=int, default=0, metavar='N',
            help='Rerun tests up to N times if they pass')
        group.add_argument(
            '--retest-until-pass',
            type=int, default=0, metavar='N',
            help='Rerun failing tests up to N times')
        parser.add_argument(
            '--abort-on-error',
            action='store_true',
            help='Abort after the first package with any errors (failing '
                 'tests are not considered errors in this context)')
        parser.add_argument(
            '--return-code-on-test-failure',
            action='store_true',
            help='Use a non-zero return code to indicate any test failure')
        add_executor_arguments(parser)
        add_event_handler_arguments(parser)

        add_packages_arguments(parser)

        decorated_parser = DestinationCollectorDecorator(parser)
        add_task_arguments(decorated_parser, 'colcon_core.task.test')
        self.task_argument_destinations = decorated_parser.get_destinations()
        self.task_argument_destinations['retest-until-pass'] = \
            'retest_until_pass'
        self.task_argument_destinations['retest-until-fail'] = \
            'retest_until_fail'

    def main(self, *, context):  # noqa: D102
        check_and_mark_build_tool(context.args.build_base)
        check_and_mark_install_layout(
            context.args.install_base,
            merge_install=context.args.merge_install)

        decorators = get_packages(
            context.args,
            additional_argument_names=self.task_argument_destinations,
            recursive_categories=('run', ))

        install_base = os.path.abspath(os.path.join(
            os.getcwd(), context.args.install_base))
        jobs = self._get_jobs(context.args, decorators, install_base)

        if context.args.return_code_on_test_failure:
            # watch published events on all jobs to detect any test failures
            any_test_failures = False

            def check_for_test_failures(put_event_into_queue):
                nonlocal any_test_failures

                def put_event_into_queue_(self, event):
                    nonlocal any_test_failures
                    nonlocal put_event_into_queue
                    if isinstance(event, TestFailure):
                        any_test_failures = True
                    return put_event_into_queue(event)

                return put_event_into_queue_

            for job in jobs.values():
                job.put_event_into_queue = types.MethodType(
                    check_for_test_failures(job.put_event_into_queue), job)

        on_error = OnError.continue_ \
            if not context.args.abort_on_error else OnError.interrupt
        rc = execute_jobs(context, jobs, on_error=on_error)

        if context.args.return_code_on_test_failure:
            if not rc and any_test_failures:
                return 1

        return rc

    def _get_jobs(self, args, decorators, install_base):
        jobs = OrderedDict()
        for decorator in decorators:
            if not decorator.selected:
                continue

            pkg = decorator.descriptor
            extension = get_task_extension('colcon_core.task.test', pkg.type)
            if not extension:
                logger.warning(
                    "No task extension to 'test' a '{pkg.type}' package"
                    .format_map(locals()))
                continue

            recursive_dependencies = OrderedDict()
            # for testing a package include itself in the environment
            for dep_name in decorator.recursive_dependencies + [pkg.name]:
                dep_path = install_base
                if not args.merge_install:
                    dep_path = os.path.join(dep_path, dep_name)
                recursive_dependencies[dep_name] = dep_path

            package_args = TestPackageArguments(
                pkg, args, additional_destinations=self
                .task_argument_destinations.values())
            ordered_package_args = ', '.join([
                ('%s: %s' % (repr(k), repr(package_args.__dict__[k])))
                for k in sorted(package_args.__dict__.keys())
            ])
            logger.debug(
                "Testing package '{pkg.name}' with the following arguments: "
                '{{{ordered_package_args}}}'.format_map(locals()))
            task_context = TaskContext(
                pkg=pkg, args=package_args,
                dependencies=recursive_dependencies)

            job = Job(
                identifier=pkg.name,
                dependencies=set(recursive_dependencies.keys()),
                task=extension, task_context=task_context)

            jobs[pkg.name] = job
        return jobs



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from asyncio import CancelledError
from enum import Enum
import inspect
import os
import traceback
import warnings

from colcon_core.environment_variable import EnvironmentVariable
from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobQueued
from colcon_core.event.job import JobSkipped
from colcon_core.event.job import JobStarted
from colcon_core.event.output import StderrLine
from colcon_core.event_reactor import create_event_reactor
from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import get_first_line_doc
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_grouped_by_priority
from colcon_core.subprocess import SIGINT_RESULT

logger = colcon_logger.getChild(__name__)

"""Environment variable to override the default executor"""
DEFAULT_EXECUTOR_ENVIRONMENT_VARIABLE = EnvironmentVariable(
    'COLCON_DEFAULT_EXECUTOR', 'Select the default executor extension')


class Job:
    """A job describes a unit of work."""

    def __init__(self, *, identifier, dependencies, task, task_context):
        """
        Construct a Job.

        :param str identifier: The job identifier
        :param set dependencies: The identifiers of other jobs which this job
          depends on
        :param task: The task extension
        :param task_context: The task context
        """
        self._event_queue = None
        self.identifier = identifier
        self.dependencies = dependencies
        self.task = task
        self.returncode = None
        self.task_context = task_context

    def set_event_queue(self, event_queue):
        """
        Set the event queue.

        Using the event queue the job can post events.
        A :class:`JobQueued` event with the package name from the task context
        is posted to the event queue.

        :param event_queue: The event queue
        """
        self._event_queue = event_queue

        self.put_event_into_queue(JobQueued(
            self.task_context.pkg.name, self.task_context.dependencies))

    async def __call__(self, *args, **kwargs):
        """
        Perform the unit of work.

        The overview of the process:
        * Put a :class:`JobStarted` event into the queue
        * Pass the task context to the task
        * Invoke the task
        * In case the task is canceled return a :attribute:`SIGINT_RESULT`
          code
        * In case of an exception within the task put a :class:`StderrLine`
          event into the queue and re-raise the exception
        * Put a :class:`JobEnded` event into the queue

        :returns: The return code of the invoked task
        :raises Exception: Any exception the invoked task raises
        """
        self.put_event_into_queue(JobStarted(self.task_context.pkg.name))

        # replace function to use this job as the event context
        self.task_context.put_event_into_queue = self.put_event_into_queue
        self.task.set_context(context=self.task_context)

        rc = 0
        try:
            rc = await self.task(*args, **kwargs)
        except CancelledError:
            rc = SIGINT_RESULT
        except Exception:
            rc = 1
            self.put_event_into_queue(
                StderrLine(traceback.format_exc().encode()))
            raise
        finally:
            if self.returncode is None:
                self.returncode = rc or 0
            self.put_event_into_queue(
                JobEnded(self.task_context.pkg.name, self.returncode))
        return self.returncode

    def put_event_into_queue(self, event):
        """
        Post a message event into the event queue.

        :param event: The event
        """
        self._event_queue.put((event, self))

    def __str__(self):
        """Use the identifier as the string representation of a job."""
        return self.identifier


class OnError(Enum):
    """Decision how to proceed when one job fails."""

    # ongoing jobs will continue, pending jobs will be executed
    continue_ = 1
    # ongoing jobs will be cancelled, pending jobs won't be executed
    interrupt = 2
    # ongoing jobs will continue, pending jobs won't be executed
    skip_pending = 3
    # ongoing jobs will continue, pending jobs will only be executed if they
    # don't (recursively) depend on a failed job
    skip_downstream = 4


class ExecutorExtensionPoint:
    """
    The interface for executor extensions.

    An executor extension runs a set of jobs.

    For each instance the attribute `EXECUTOR_NAME` is being set to the
    basename of the entry point registering the extension.
    """

    """The version of the executor extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of executor extensions."""
    PRIORITY = 100

    def __init__(self):  # noqa: D107
        super().__init__()
        self._event_controller = None

    def add_arguments(self, *, parser):
        """
        Add command line arguments specific to the executor.

        The method is intended to be overridden in a subclass.

        :param parser: The argument parser
        """
        pass

    def set_event_controller(self, event_controller):
        """
        Set the event controller.

        Using the event controller the executor can force a flush of all
        events.
        """
        self._event_controller = event_controller

    def execute(
        self, args, jobs, *, on_error: OnError = None, abort_on_error=None
    ):
        """
        Execute the passed jobs.

        This method must be overridden in a subclass.
        Subclass should not include the deprecated keyword argument
        `abort_on_error` in their signature.

        :param arguments: The passed arguments

        The deprecated API accepts the following separate arguments:
        :param args: The parsed command line arguments
        :param jobs: The jobs
        :param on_error: The decision how to proceed when one job fails
        :param abort_on_error: The flag if pending jobs should be aborted in
          case of any errors or individual jobs failing (deprecated, use
          `on_error` instead)
        """
        raise NotImplementedError()

    def _flush(self):
        if self._event_controller is None:
            return
        self._event_controller.flush()


def get_executor_extensions():
    """
    Get the available executor extensions.

    The extensions are grouped by their priority and each group is ordered by
    the entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.EXECUTOR_NAME = name
    return order_extensions_grouped_by_priority(extensions)


def add_executor_arguments(parser):
    """
    Add the command line arguments for the executor extensions.

    :param parser: The argument parser
    """
    group = parser.add_argument_group(title='Executor arguments')
    extensions = get_executor_extensions()
    keys = []
    descriptions = ''
    for priority in extensions.keys():
        extensions_same_prio = extensions[priority]
        assert len(extensions_same_prio) == 1, \
            'Executor extensions must have unique priorities'
        for key, extension in extensions_same_prio.items():
            keys.append(key)
            desc = get_first_line_doc(extension)
            if not desc:
                # show extensions without a description
                # to mention the available options
                desc = '<no description>'
            # it requires a custom formatter to maintain the newline
            descriptions += '\n* {key}: {desc}'.format_map(locals())
    assert keys, 'No executor extensions found'

    default = os.environ.get(DEFAULT_EXECUTOR_ENVIRONMENT_VARIABLE.name)
    if default not in keys:
        default = keys[0]

    group.add_argument(
        '--executor', type=str, choices=keys, default=default,
        help='The executor to process all packages (default: {default})'
             '{descriptions}'.format_map(locals()))

    for priority in extensions.keys():
        extensions_same_prio = extensions[priority]
        for extension in extensions_same_prio.values():
            try:
                retval = extension.add_arguments(parser=group)
                assert retval is None, 'add_arguments() should return None'
            except Exception as e:  # noqa: F841
                # catch exceptions raised in executor extension
                exc = traceback.format_exc()
                logger.error(
                    'Exception in executor extension '
                    "'{extension.EXECUTOR_NAME}': {e}\n{exc}"
                    .format_map(locals()))
                # skip failing extension, continue with next one


def execute_jobs(
    context, jobs, *, on_error: OnError = None, abort_on_error=None
):
    """
    Execute jobs.

    The overview of the process:
    * One executor extension is being chosen based on the command line
      arguments.
    * Create an event controller.
    * Pass the event controller to the executor extension.
    * Pass the event queue to all jobs.
    * Start the event controller.
    * Invoke the executor extension to execute the jobs.
    * Join the event controller.

    :param jobs: The ordered dictionary of jobs
    :param on_error: The decision how to proceed when one job fails
    :param abort_on_error: The flag if pending jobs should be aborted in case
      of individual jobs failing (deprecated, use `on_error` instead)
    :returns: The return code
    """
    assert on_error is None or abort_on_error is None, \
        'Only one of the two keyword arguments can be passed'
    # keep default behavior of deprecated keyword argument
    if on_error is None and abort_on_error is None:
        on_error = OnError.interrupt
    if abort_on_error is not None:
        warnings.warn(
            "'colcon_core.executor.execute_jobs' was called with the "
            "deprecated keyword argument 'abort_on_error'", stacklevel=2)
        on_error = OnError.interrupt if abort_on_error else OnError.continue_

    executor = select_executor_extension(context.args)
    assert executor

    logger.info("Executing jobs using '%s' executor", executor.EXECUTOR_NAME)

    # create event reactor with handlers specified by the args
    with create_event_reactor(context) as event_controller:
        executor.set_event_controller(event_controller)

        # pass queue to jobs to publish events
        for job in jobs.values():
            job.set_event_queue(event_controller.get_queue())

        func = executor.execute
        signature = inspect.signature(func)
        kwargs = {}
        if 'on_error' in signature.parameters:
            kwargs['on_error'] = on_error
        else:
            # fallback to legacy API
            assert 'abort_on_error' in signature.parameters
            warnings.warn(
                "The ExecutorExtensionPoint '{executor.EXECUTOR_NAME}' uses a "
                "deprecated signature for the 'execute' method"
                .format_map(locals()))
            kwargs['abort_on_error'] = on_error == OnError.interrupt

        try:
            rc = func(context.args, jobs, **kwargs)
        except Exception as e:  # noqa: F841
            # catch exceptions raised in executor extension
            exc = traceback.format_exc()
            logger.error(
                "Exception in executor extension '{executor.EXECUTOR_NAME}': "
                '{e}\n{exc}'.format_map(locals()))
            rc = 1
        finally:
            # generate an event for every skipped job
            for job in jobs.values():
                if job.returncode is not None:
                    continue
                event_controller.get_queue().put(
                    (JobSkipped(job.identifier), job))

    return rc


def select_executor_extension(args):
    """
    Get the executor extension.

    :param args: The parsed command line arguments
    :returns: The executor extension
    """
    executor_extensions = get_executor_extensions()
    for priority in executor_extensions.keys():
        extensions_same_prio = executor_extensions[priority]
        for key, extension in extensions_same_prio.items():
            if key == args.executor:
                return extension
    # one executor should always be selected by the default value
    # in case their are no executor extensions available the add argument
    # function should have already failed
    assert False



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.package_descriptor import PackageDescriptor
from colcon_core.package_discovery import _discover_packages
from colcon_core.package_discovery import _get_extensions_with_parameters
from colcon_core.package_discovery import add_package_discovery_arguments
from colcon_core.package_discovery import discover_packages
from colcon_core.package_discovery import get_package_discovery_extensions
from colcon_core.package_discovery import PackageDiscoveryExtensionPoint
from mock import Mock
from mock import patch

from .entry_point_context import EntryPointContext


class Extension1(PackageDiscoveryExtensionPoint):
    PRIORITY = 80


class Extension2(PackageDiscoveryExtensionPoint):
    pass


class Extension3(PackageDiscoveryExtensionPoint):
    PRIORITY = 90


class Extension4(PackageDiscoveryExtensionPoint):
    pass


def test_get_package_discovery_extensions():
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        extensions = get_package_discovery_extensions()
        assert ['extension2', 'extension4', 'extension3', 'extension1'] == \
            list(extensions.keys())


def test_add_package_discovery_arguments():
    parser = Mock()
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        add_package_discovery_arguments(parser)

        all_extensions = get_package_discovery_extensions()

        extensions = {'extension2': all_extensions['extension2']}
        # mock the has_default method
        extensions['extension2'].has_default = Mock(
            side_effect=ValueError('exception in has_default'))
        with patch('colcon_core.package_discovery.logger.error') as error:
            add_package_discovery_arguments(parser, extensions=extensions)
            # the raised exception is catched and results in an error message
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in package discovery extension 'extension2': "
                'exception in has_default\n')

        extensions = {'extension3': all_extensions['extension3']}
        # mock the add_arguments method
        extensions['extension3'].add_arguments = Mock(
            side_effect=ValueError('exception in add_arguments'))
        with patch('colcon_core.package_discovery.logger.error') as error:
            add_package_discovery_arguments(parser, extensions=extensions)
            # the raised exception is catched and results in an error message
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in package discovery extension 'extension3': "
                'exception in add_arguments\n')

        # mock the has_default method and return True
        extensions['extension3'].has_default = Mock(return_value=True)
        add_package_discovery_arguments(parser, extensions=extensions)


def test_discover_packages():
    # check without any extensions
    with patch(
        'colcon_core.package_discovery.get_package_discovery_extensions',
        return_value={},
    ) as get_extensions:
        with patch('colcon_core.package_discovery.logger.warning') as warn:
            descs = discover_packages(None, None)
    assert get_extensions.call_count == 1
    warn.assert_called_once_with('No package discovery extensions found')
    assert descs == set()

    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        extensions = get_package_discovery_extensions()
        assert len(extensions) == 4

        # check without any parameters
        extensions['extension1'].discover = Mock(
            return_value={PackageDescriptor('/extension1/pkg1')})
        extensions['extension2'].discover = Mock(
            return_value={PackageDescriptor('/extension2/pkg1')})

        descs = discover_packages(None, None, discovery_extensions=extensions)
        assert len(descs) == 2
        expected_path = '/extension1/pkg1'.replace('/', os.sep)
        assert expected_path in {str(d.path) for d in descs}
        expected_path = '/extension2/pkg1'.replace('/', os.sep)
        assert expected_path in {str(d.path) for d in descs}

        # check with parameters
        extensions['extension3'].has_parameters = Mock(return_value=True)
        extensions['extension3'].discover = Mock(
            return_value={
                PackageDescriptor('/extension3/pkg1'),
                PackageDescriptor('/extension3/pkg2')})

        descs = discover_packages(None, None, discovery_extensions=extensions)
        assert len(descs) == 2
        expected_path = '/extension3/pkg1'.replace('/', os.sep)
        assert expected_path in {str(d.path) for d in descs}
        expected_path = '/extension3/pkg2'.replace('/', os.sep)
        assert expected_path in {str(d.path) for d in descs}


def test__get_extensions_with_parameters():
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        extensions = get_package_discovery_extensions()

        # mock the has_parameters method
        extensions['extension1'].has_parameters = Mock(return_value=True)
        extensions['extension2'].has_parameters = Mock(
            side_effect=ValueError('exception in has_parameters'))
        extensions['extension3'].has_parameters = Mock(
            side_effect=PackageDiscoveryExtensionPoint.has_parameters)
        extensions['extension4'].has_parameters = Mock(return_value=False)

        with_parameters = _get_extensions_with_parameters(Mock(), extensions)
        assert extensions['extension1'].has_parameters.call_count == 1
        assert extensions['extension2'].has_parameters.call_count == 1
        assert extensions['extension3'].has_parameters.call_count == 1
        assert extensions['extension4'].has_parameters.call_count == 1
        assert set(with_parameters.keys()) == {'extension1'}


def test__discover_packages():
    descs = _discover_packages(None, None, {})
    assert descs == set()

    with EntryPointContext(
        extension1=Extension1, extension2=Extension2,
        extension3=Extension3, extension4=Extension4,
    ):
        extensions = get_package_discovery_extensions()

        # mock the discover method in the order the extensions are being called
        extensions['extension2'].discover = Mock(
            side_effect=ValueError('exception in discover'))
        extensions['extension4'].discover = Mock(
            side_effect=extensions['extension4'].discover)
        extensions['extension3'].discover = Mock(
            return_value={
                PackageDescriptor('/extension3/pkg1'),
                PackageDescriptor('/extension3/pkg2')})
        # returns None instead of a set
        extensions['extension1'].discover = Mock()

        with patch('colcon_core.package_discovery.logger.error') as error:
            descs = _discover_packages(Mock(), None, extensions)

            # in the order the extensions are being called
            assert extensions['extension2'].discover.call_count == 1
            assert extensions['extension4'].discover.call_count == 1
            assert extensions['extension3'].discover.call_count == 1
            assert extensions['extension1'].discover.call_count == 1

            # the raised exceptions are catched and result in error messages
            assert error.call_count == 2
            assert len(error.call_args_list[0][0]) == 1
            assert error.call_args_list[0][0][0].startswith(
                "Exception in package discovery extension 'extension2': "
                'exception in discover\n')
            assert len(error.call_args_list[1][0]) == 1
            assert error.call_args_list[1][0][0].startswith(
                "Exception in package discovery extension 'extension1': "
                'discover() should return a set\n')

        assert len(descs) == 2
        expected_path = '/extension3/pkg1'.replace('/', os.sep)
        assert expected_path in {str(d.path) for d in descs}
        expected_path = '/extension3/pkg2'.replace('/', os.sep)
        assert expected_path in {str(d.path) for d in descs}



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from sys import executable

from colcon_core.plugin_system import satisfies_version
from colcon_core.task import check_call
from colcon_core.task.python.test import PythonTestingStepExtensionPoint
from colcon_core.verb.test import logger


class SetuppyPythonTestingStep(PythonTestingStepExtensionPoint):
    """Use `setup.py test` to test packages."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            PythonTestingStepExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

    def match(self, context, env, setup_py_data):  # noqa: D102
        return True

    async def step(self, context, env, setup_py_data):  # noqa: D102
        if context.args.retest_until_fail:
            logger.warning(
                "Ignored '--retest-until-fail' for package "
                "'{context.pkg.name}' since 'setup.py test' does not support "
                'the usage'.format_map(locals()))

        if context.args.retest_until_pass:
            logger.warning(
                "Ignored '--retest-until-pass' for package "
                "'{context.pkg.name}' since 'setup.py test' does not support "
                'the usage'.format_map(locals()))

        cmd = [
            executable,
            'setup.py', 'test',
            'egg_info', '--egg-base', context.args.build_base,
        ]
        rc = await check_call(context, cmd, cwd=context.args.path, env=env)
        if rc and rc.returncode:
            return rc.returncode



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from argparse import ArgumentParser
from asyncio import CancelledError

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobQueued
from colcon_core.event.job import JobSkipped
from colcon_core.event.job import JobStarted
from colcon_core.event.output import StderrLine
from colcon_core.executor import add_executor_arguments
from colcon_core.executor import DEFAULT_EXECUTOR_ENVIRONMENT_VARIABLE
from colcon_core.executor import execute_jobs
from colcon_core.executor import ExecutorExtensionPoint
from colcon_core.executor import get_executor_extensions
from colcon_core.executor import Job
from colcon_core.executor import OnError
from colcon_core.subprocess import SIGINT_RESULT
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext
from .environment_context import EnvironmentContext
from .run_until_complete import run_until_complete


class Task:

    def __init__(self):
        self.return_value = None

    def set_context(self, *, context):
        pass

    async def __call__(self, *args, **kwargs):
        if isinstance(self.return_value, BaseException):
            raise self.return_value
        return self.return_value


def test_job():
    task = Task()
    task_context = Mock()
    task_context.dependencies = Mock()
    task_context.pkg = Mock()
    task_context.pkg.name = 'name'
    job = Job(
        identifier='id', dependencies=set(), task=task,
        task_context=task_context)
    assert str(job) == 'id'

    events = []
    event_queue = Mock()
    event_queue.put = lambda event: events.append(event)
    job.set_event_queue(event_queue)
    assert len(events) == 1
    assert isinstance(events[-1][0], JobQueued)
    assert events[-1][0].identifier == 'name'
    assert events[-1][0].dependencies == task_context.dependencies
    assert events[-1][1] == job

    # successful task
    rc = run_until_complete(job())
    assert rc == 0
    assert len(events) == 3
    assert isinstance(events[-2][0], JobStarted)
    assert events[-2][0].identifier == 'name'
    assert events[-2][1] == job
    assert isinstance(events[-1][0], JobEnded)
    assert events[-1][0].identifier == 'name'
    assert events[-1][0].rc == 0
    assert events[-1][1] == job

    # canceled task
    job.returncode = None
    task.return_value = CancelledError()
    rc = run_until_complete(job())
    assert rc is SIGINT_RESULT
    assert len(events) == 5
    assert isinstance(events[-2][0], JobStarted)
    assert events[-2][0].identifier == 'name'
    assert events[-2][1] == job
    assert isinstance(events[-1][0], JobEnded)
    assert events[-1][0].identifier == 'name'
    assert events[-1][0].rc is SIGINT_RESULT
    assert events[-1][1] == job

    # task raising exception
    job.returncode = None
    task.return_value = RuntimeError('custom exception')
    with pytest.raises(RuntimeError):
        run_until_complete(job())
    assert len(events) == 8
    assert isinstance(events[-3][0], JobStarted)
    assert events[-3][0].identifier == 'name'
    assert events[-3][1] == job
    assert isinstance(events[-2][0], StderrLine)
    assert events[-2][0].line.endswith(b'\nRuntimeError: custom exception\n')
    assert events[-2][1] == job
    assert isinstance(events[-1][0], JobEnded)
    assert events[-1][0].identifier == 'name'
    assert events[-1][0].rc == 1
    assert events[-1][1] == job

    # override task return code
    job.returncode = 2
    task.return_value = 0
    rc = run_until_complete(job())
    assert rc == 2
    assert len(events) == 10
    assert isinstance(events[-2][0], JobStarted)
    assert events[-2][0].identifier == 'name'
    assert events[-2][1] == job
    assert isinstance(events[-1][0], JobEnded)
    assert events[-1][0].identifier == 'name'
    assert events[-1][0].rc == 2
    assert events[-1][1] == job


def test_interface():
    interface = ExecutorExtensionPoint()
    interface._flush()
    event_controller = Mock()
    interface.set_event_controller(event_controller)
    interface._flush()
    assert event_controller.flush.call_count == 1


class Extension1(ExecutorExtensionPoint):
    """Class documentation."""


class Extension2(ExecutorExtensionPoint):
    PRIORITY = 110


class Extension3(ExecutorExtensionPoint):
    pass


def test_add_executor_arguments():
    parser = ArgumentParser()
    # extensions with the same priority
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2, extension3=Extension3
    ):
        with pytest.raises(AssertionError) as e:
            add_executor_arguments(parser)
        assert 'Executor extensions must have unique priorities' in \
            str(e.value)

    # no extensions
    with EntryPointContext():
        with pytest.raises(AssertionError) as e:
            add_executor_arguments(parser)
        assert 'No executor extensions found' in str(e.value)

    # choose executor by environment variable
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        extensions = get_executor_extensions()
        extensions[110]['extension2'].add_arguments = Mock(
            side_effect=RuntimeError('custom exception'))
        extensions[100]['extension1'].add_arguments = Mock(return_value=None)
        env = {DEFAULT_EXECUTOR_ENVIRONMENT_VARIABLE.name: 'extension1'}
        with EnvironmentContext(**env):
            with patch('colcon_core.executor.logger.error') as error:
                add_executor_arguments(parser)
    assert extensions[100]['extension1'].add_arguments.call_count == 1
    # the raised exception is catched and results in an error message
    assert error.call_count == 1
    assert len(error.call_args[0]) == 1
    assert error.call_args[0][0].startswith(
        "Exception in executor extension 'extension2': custom exception\n")
    args = parser.parse_args([])
    assert args.executor == 'extension1'

    # choose default executor
    parser = ArgumentParser()
    with EntryPointContext(extension1=Extension1, extension2=Extension2):
        add_executor_arguments(parser)
    args = parser.parse_args([])
    assert args.executor == 'extension2'


def test_execute_jobs():
    context = Mock()
    context.args = Mock()
    context.args.event_handlers = None
    task_context = Mock()
    task_context.pkg = Mock()
    task_context.pkg.name = 'name'
    jobs = {
        'one': Job(
            identifier='id', dependencies=set(), task=None,
            task_context=task_context)}

    event_reactor = Mock()
    event_reactor.__enter__ = lambda self: self
    event_reactor.__exit__ = lambda self, *args: None
    with patch(
        'colcon_core.executor.create_event_reactor', return_value=event_reactor
    ):
        with EntryPointContext(extension1=Extension1, extension2=Extension2):
            # no extension selected
            with pytest.raises(AssertionError):
                execute_jobs(context, jobs)

            # execute method not implemented and sending skipped job event
            context.args.executor = 'extension2'
            with patch('colcon_core.executor.logger.error') as error:
                rc = execute_jobs(context, jobs)
            assert rc == 1
            assert error.call_count == 1
            assert len(error.call_args[0]) == 1
            assert error.call_args[0][0].startswith(
                "Exception in executor extension 'extension2': \n")
            assert event_reactor.get_queue().put.call_count == 2
            assert isinstance(
                event_reactor.get_queue().put.call_args_list[0][0][0][0],
                JobQueued)
            assert isinstance(
                event_reactor.get_queue().put.call_args_list[1][0][0][0],
                JobSkipped)

            # successful execution
            event_reactor.get_queue().put.reset_mock()
            jobs['one'].returncode = 0
            extensions = get_executor_extensions()
            extensions[110]['extension2'].execute = \
                lambda args, jobs, on_error: 0
            rc = execute_jobs(context, jobs, on_error=OnError.interrupt)
            assert rc == 0
            assert event_reactor.get_queue().put.call_count == 1
            assert isinstance(
                event_reactor.get_queue().put.call_args[0][0][0], JobQueued)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.package_decorator import add_recursive_dependencies
from colcon_core.package_decorator import get_decorators
from colcon_core.package_decorator import PackageDecorator
from colcon_core.package_descriptor import PackageDescriptor
from mock import Mock


def test_constructor():
    desc = Mock()
    d = PackageDecorator(desc)
    assert d.descriptor == desc
    assert d.recursive_dependencies is None
    assert d.selected is True


def test_get_decorators():
    desc1 = Mock()
    desc2 = Mock()
    decos = get_decorators([desc1, desc2])
    assert len(decos) == 2
    assert decos[0].descriptor == desc1
    assert decos[1].descriptor == desc2


def test_add_recursive_dependencies():
    d = PackageDescriptor('/some/path')
    d.name = 'A'
    d.dependencies['build'].add('B')
    d.dependencies['build'].add('c')
    d.dependencies['run'].add('D')
    d.dependencies['test'].add('e')

    d1 = PackageDescriptor('/other/path')
    d1.name = 'B'
    d1.dependencies['build'].add('f')
    d1.dependencies['run'].add('G')

    d2 = PackageDescriptor('/other/path')
    d2.name = 'D'
    d2.dependencies['run'].add('h')

    d3 = PackageDescriptor('/another/path')
    d3.name = 'G'
    d3.dependencies['build'].add('i')

    decos = get_decorators([d, d1, d2, d3])
    add_recursive_dependencies(
        decos, direct_categories={'build', 'run'},
        recursive_categories={'run'})

    assert decos[0].recursive_dependencies is not None
    assert decos[1].recursive_dependencies is not None
    assert decos[2].recursive_dependencies is not None
    assert decos[3].recursive_dependencies is not None
    assert decos[0].recursive_dependencies == {'B', 'D', 'G'}



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import shutil
import signal
import sys
from tempfile import mkdtemp

from colcon_core.command import CommandContext
from colcon_core.command import create_parser
from colcon_core.command import main
from colcon_core.command import verb_main
from colcon_core.environment_variable import EnvironmentVariable
from colcon_core.verb import VerbExtensionPoint
from mock import Mock
from mock import patch
import pytest

from .entry_point_context import EntryPointContext


class Extension1(VerbExtensionPoint):
    pass


class Extension2:
    """Very long line so that the help text needs to be wrapped."""

    def main(self, *, context):
        pass  # pragma: no cover


class Extension3(VerbExtensionPoint):

    def add_arguments(self, *, parser):
        raise RuntimeError('custom exception')


def test_main():
    with EntryPointContext(
        extension1=Extension1, extension2=Extension2, extension3=Extension3
    ):
        with patch(
            'colcon_core.argument_parser.get_argument_parser_extensions',
            return_value={}
        ):
            with pytest.raises(SystemExit) as e:
                main(argv=['--help'])
            assert e.value.code == 0

            with pytest.raises(SystemExit) as e:
                main(argv=['--log-level', 'invalid'])
            assert e.value.code == 2

            # avoid creating log directory in the package directory
            log_base = mkdtemp(prefix='test_colcon_')
            argv = ['--log-base', log_base]
            try:
                main(argv=argv + ['--log-level', 'info'])

                with patch(
                    'colcon_core.command.load_entry_points',
                    return_value={
                        'key1': EnvironmentVariable('name', 'description'),
                        'key2': EnvironmentVariable(
                            'extra_long_name_to_wrap help',
                            'extra long description text to require a wrap of '
                            'the help text not_only_on_spaces_but_also_forced_'
                            'within_a_very_long_consecutive_word'),
                    }
                ):
                    main(argv=argv + ['extension1'])
            finally:
                # the logging subsystem might still have file handles pending
                # therefore only try to delete the temporary directory
                shutil.rmtree(log_base, ignore_errors=True)

        # catch KeyboardInterrupt and return SIGINT error code
        with patch('colcon_core.command._main', return_value=0) as _main:
            _main.side_effect = KeyboardInterrupt()
            rc = main()
            assert rc == signal.SIGINT


def test_create_parser():
    with EntryPointContext():
        parser = create_parser('colcon_core.environment_variable')

    parser.add_argument('--foo', nargs='*', type=str.lstrip)
    args = parser.parse_args(['--foo', '--bar', '--baz'])
    assert args.foo == ['--bar', '--baz']

    parser.add_argument('--baz', action='store_true')
    args = parser.parse_args(['--foo', '--bar', '--baz'])
    assert args.foo == ['--bar']
    assert args.baz is True

    args = parser.parse_args(['--foo', '--bar', ' --baz'])
    assert args.foo == ['--bar', '--baz']

    argv = sys.argv
    sys.argv = ['/some/path/prog_name/__main__.py'] + sys.argv[1:]
    with EntryPointContext():
        parser = create_parser('colcon_core.environment_variable')
    sys.argv = argv
    assert parser.prog == 'prog_name'


class Object(object):
    pass


def test_verb_main():
    args = Object()
    args.verb_name = 'verb_name'
    logger = Object()
    logger.error = Mock()

    # pass through return code
    args.main = Mock(return_value=42)
    context = CommandContext(command_name='command_name', args=args)
    rc = verb_main(context, logger)
    assert rc == args.main.return_value
    logger.error.assert_not_called()

    # catch RuntimeError and output error message
    args.main.side_effect = RuntimeError('known error condition')
    rc = verb_main(context, logger)
    assert rc
    logger.error.assert_called_once_with(
        'command_name verb_name: known error condition')
    logger.error.reset_mock()

    # catch Exception and output error message including traceback
    args.main.side_effect = Exception('custom error message')
    rc = verb_main(context, logger)
    assert rc
    assert logger.error.call_count == 1
    assert len(logger.error.call_args[0]) == 1
    assert logger.error.call_args[0][0].startswith(
        'command_name verb_name: custom error message\n')
    assert 'Exception: custom error message' in logger.error.call_args[0][0]



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from pathlib import Path
from tempfile import TemporaryDirectory

from colcon_core.package_identification import IgnoreLocationException
from colcon_core.package_identification.ignore import IGNORE_MARKER
from colcon_core.package_identification.ignore \
    import IgnorePackageIdentification
from mock import Mock
import pytest


def test_identify():
    extension = IgnorePackageIdentification()
    metadata = Mock()
    with TemporaryDirectory(prefix='test_colcon_') as basepath:
        metadata.path = Path(basepath)
        assert extension.identify(metadata) is None

        (metadata.path / IGNORE_MARKER).write_text('')
        with pytest.raises(IgnoreLocationException):
            extension.identify(metadata)
